# -*- coding: utf-8 -*-
# (c) 2020-2023 Martin Wendt and contributors; see https://github.com/mar10/stressor
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
from pprint import pformat
from textwrap import dedent

from stressor.plugins.base import (
    ActivityBase,
    ActivityCompileError,
    ScriptActivityError,
)
from stressor.util import NO_DEFAULT, check_arg, logger, shorten_string


class RunScriptActivity(ActivityBase):
    _mandatory_args = None
    _known_args = {"export", "path", "script"}
    _info_args = ("name", "path")

    def __init__(self, config_manager, **activity_args):
        """"""
        super().__init__(config_manager, **activity_args)

        path = activity_args.get("path")
        script = activity_args.get("script")
        # Allow to pass `export: null` to define 'no export wanted'
        # (omitting the argumet is considered 'undefined' and will emit a
        # warning if the script produces variables)
        export = activity_args.get("export", NO_DEFAULT)
        if export in (None, False):
            export = tuple()
        elif export is NO_DEFAULT:
            export = None

        check_arg(path, str, or_none=True)
        check_arg(script, str, or_none=True)
        check_arg(export, (str, list, tuple), or_none=True)

        if path:
            if script:
                raise ActivityCompileError(
                    "`path` and `script` args are mutually exclusive"
                )
            path = config_manager.resolve_path(path)
            with open(path, "rt") as f:
                script = f.read()
            #:
            self.script = compile(script, path, "exec")
        elif script:
            # script = dedent(self.script)
            self.script = compile(script, "<string>", "exec")
        else:
            raise ActivityCompileError("Either `path` or `script` expected")

        #: Store a shortened code snippet for debug output
        self.source = shorten_string(dedent(script), 500, 100)
        # print(self.source)

        if export is None:
            self.export = None
        elif isinstance(export, str):
            self.export = set((export,))
        else:
            self.export = set(export)
        return

    def execute(self, session, **expanded_args):
        """"""
        global_vars = {
            # "foo": 41,
            # "__builtins__": {},
        }
        # local_vars = session.context
        local_vars = session.context.copy()
        assert "result" not in local_vars
        assert "session" not in local_vars
        local_vars["session"] = session.make_session_helper()

        # prev_local_keys = set(locals())
        prev_global_keys = set(globals())
        prev_context_keys = set(local_vars.keys())

        try:
            exec(self.script, global_vars, local_vars)
        except ConnectionError as e:
            # TODO: more requests-exceptions?
            msg = "Script failed: {!r}: {}".format(e, e)
            raise ScriptActivityError(msg)
        except Exception as e:
            msg = "Script failed: {!r}: {}".format(e, e)
            if session.verbose >= 4:
                logger.exception(msg)
                raise ScriptActivityError(msg) from e
            raise ScriptActivityError(msg)
        finally:
            local_vars.pop("session")

        result = local_vars.pop("result", None)

        context_keys = set(local_vars.keys())
        new_keys = context_keys.difference(prev_context_keys)

        if new_keys:
            if self.export is None:
                logger.info(
                    "Skript activity has no `export` defined. Ignoring new variables: '{}'".format(
                        "', '".join(new_keys)
                    )
                )
            else:
                for k in self.export:
                    v = local_vars.get(k)
                    assert type(v) in (int, float, str, list, dict)
                    session.context[k] = v
                    logger.debug("Set context.{} = {!r}".format(k, v))
                # store_keys = new_keys.intersection(self.export)

        # TODO: this cannot happen?
        new_globals = set(globals().keys()).difference(prev_global_keys)
        if new_globals:
            logger.warning("Script-defined globals: {}".format(new_globals))
            raise ScriptActivityError("Script introduced globals")

        # new_context = context_keys.difference(prev_context_keys)
        # logger.info("Script-defined context-keys: {}".format(new_context))

        # new_locals = set(locals().keys()).difference(prev_local_keys)
        # if new_locals:
        #     logger.info("Script-defined locals: {}".format(new_locals))

        # logger.info("Script locals:\n{}".format(pformat(local_vars)))
        if expanded_args.get("debug") or session.verbose >= 5:
            logger.info(
                "{} {}\n  Context after execute:\n    {}\n  return value: {!r}".format(
                    session.context_stack,
                    self,
                    pformat(session.context, indent=4),
                    result,
                )
            )
        elif session.verbose >= 3 and result is not None:
            logger.info(
                "{} returnd: {!r}".format(
                    session.context_stack,
                    shorten_string(result, 200) if isinstance(result, str) else result,
                )
            )

        return result



# -*- coding: utf-8 -*-
# (c) 2020-2023 Martin Wendt and contributors; see https://github.com/mar10/stressor
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
from copy import deepcopy

from stressor.util import check_arg, get_dict_attr


class RunContext:
    """Basically a dict, holding context variables as key/value pairs.

    The :class:`ContextStack` organizes instances of this class as stack,
    so we can have a scope-dependant context.

    Values may contain `$...` macros.

    Attributes:
        parent (:class:`RunContext`):
            The stack frame parent or None if this is the root.
        name (str):
            A short name for this context. The context manager uses it to concatenate
            a path string for the current scope.
        own_attributes (dict):
            A dict of attributes that are explicitly defined by this instance.
        all_attributes (dict):
            The aggregated context, created by overloading the parent's context
            with our own ``attributes``
    """

    def __init__(self, parent, name, update_attributes=None, copy_data=False):
        """
        Args:
            parent (:class:`RunContext`):
                The stack frame parent or None if this is the root.
            name (str):
                A short name for this context. The context manager uses it to concatenate
                a path string for the current scope.
            update_attributes (dict):
                A dict of attributes that are explicitly defined by this instance.
            copy_data (bool):

        """
        check_arg(parent, RunContext, or_none=True)
        check_arg(name, str, name != "")
        check_arg(update_attributes, dict, or_none=True)
        check_arg(copy_data, bool)

        self.parent = parent
        self.name = name
        if update_attributes is None:
            self.own_attributes = {}
        else:
            self.own_attributes = deepcopy(update_attributes)

        if parent is None:
            self.all_attributes = self.own_attributes
        elif copy_data:
            # Create a copy of the parent's context, so the original state is
            # restored on pull
            self.all_attributes = deepcopy(parent.all_attributes)
            if self.own_attributes:
                self.all_attributes.update(self.own_attributes)
        else:
            # We only reference the parent context instance, so change will persist
            # after a popping from this stack
            self.all_attributes = parent.all_attributes

        return


class ContextStack:
    """
    The context manager
    Examples:
    """

    MAX_DEPTH = 100

    def __init__(self, name=None, context=None):
        self.ctx_stack = []
        if context is not None:
            assert name
            rc = RunContext(None, name or "root", context)
            self._push(rc)

    def __str__(self):
        return self.path()

    #: Provide nicer display for pprint(), etc.
    def __repr__(self):
        return f"ContextStack<{self}>"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.pop()
        return

    @property
    def context(self):
        ctx = self.ctx_stack[-1].all_attributes
        return ctx

    def enter(self, name, attributes=None):
        self.push(name, attributes)
        return self

    def level(self):
        return len(self.ctx_stack)

    def _push(self, context):
        """
        Raises:
            RuntimeError if queue length exceeds ContextStack.MAX_DEPTH
        """
        check_arg(context, RunContext)

        if len(self.ctx_stack) >= self.MAX_DEPTH:
            raise RuntimeError("Max depth exceeded ({})".format(self.MAX_DEPTH))
        self.ctx_stack.append(context)
        return context

    def push(self, name, attributes=None, copy_data=False):
        """Push `name` to the stack and optionally update or copy context.
        Args:
            name (str):
        Raises:
            RuntimeError if queue length exceeds ContextStack.MAX_DEPTH
        """
        check_arg(name, str)
        check_arg(attributes, dict, or_none=True)

        try:
            parent = self.peek()
        except IndexError:
            parent = None

        ctx = RunContext(parent, name, attributes, copy_data)
        return self._push(ctx)

    def pop(self):
        """
        Raises:
            IndexError if queue is empty
        """
        ctx = self.ctx_stack.pop()
        return ctx

    def peek(self, offset=1):
        """
        Raises:
            IndexError if offset is invalid
        """
        assert offset > 0
        ctx = self.ctx_stack[-offset]
        return ctx

    def path(self):
        path = "/".join((ctx.name for ctx in self.ctx_stack))
        return "/" + path

    def as_dict(self):
        """Return the current aggregated context."""
        return self.context

    def get_attr(self, key_path, context=None):
        check_arg(key_path, str)
        check_arg(context, RunContext, or_none=True)
        if context is None:
            context = self.peek()
        return get_dict_attr(self.as_dict(), key_path)

    def set_last_part(self, name):
        self.ctx_stack[-1].name = name



# -*- coding: utf-8 -*-
# (c) 2020-2023 Martin Wendt and contributors; see https://github.com/mar10/stressor
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
import re
import threading
from pprint import pformat
from queue import Empty, Queue
from urllib.parse import urlencode

import requests
from lxml import html
from requests.exceptions import RequestException

from stressor import __version__
from stressor.plugins.base import (
    ActivityAssertionError,
    ActivityBase,
    ActivityError,
    ActivityTimeoutError,
)
from stressor.util import (
    check_arg,
    get_dict_attr,
    is_relative_url,
    logger,
    resolve_url,
    shorten_string,
)


def match_value(pattern, value, info):
    """Return ."""
    check_arg(pattern, str)

    value = str(value)
    match = re.match(pattern, value)  # , re.MULTILINE)
    # if "." in pattern or "*" in pattern:
    #     match = re.match(pattern, value)  # , re.MULTILINE)
    # else:
    #     match = pattern == value

    if not match:
        msg = "`{}` value {!r} does not match pattern {!r}".format(info, value, pattern)
        return False, msg
    return True, None


class HTTPRequestActivity(ActivityBase):
    # RESPONSE_LOG_LENGTH = 200
    REQUEST_ARGS = {"auth", "data", "json", "headers", "params", "timeout", "verify"}
    _mandatory_args = {"method", "url"}
    _known_args = (
        REQUEST_ARGS
        | _mandatory_args
        | {
            "store_json",
            "assert_match_headers",
            "assert_status",
            "assert_json",
            "assert_html",
        }
    )

    _info_args = ("name", "method", "url")

    def __init__(self, config_manager, **activity_args):
        super().__init__(config_manager, **activity_args)

        check_arg(activity_args.get("method"), str)
        check_arg(activity_args.get("url"), str)
        check_arg(activity_args.get("params"), dict, or_none=True)

        # self.r_args = {k: v for k, v in self.raw_args.items() if k in self.REQUEST_ARGS}
        # self.a_args = {
        #     k: v for k, v in self.raw_args.items() if k not in self.REQUEST_ARGS
        # }

    def get_info(self, info_args=True, expanded_args=None, session=None):
        args_dict = expanded_args if expanded_args else self.raw_args
        url = args_dict.get("url")
        params = args_dict.get("params") or ""

        if params:
            params = "?" + urlencode(params)
        if self.__class__ is HTTPRequestActivity:
            method = "{} ".format(args_dict["method"])
        else:
            method = ""
        return "{}({}{}{})".format(self.get_script_name(), method, url, params)

    @classmethod
    def _format_response(cls, resp, short=False, add_headers=False, ruler=True):

        try:
            s = resp.json()
            s = pformat(s)
        except ValueError:
            rt = resp.text
            lines = []
            for s in rt.split("\n"):
                s = s.rstrip()
                if s:
                    lines.append(s)
            s = "\n".join(lines)
            # s = fill(s)

        res = []
        res.append("")
        res.append(
            "--- Response status: {}, len: {:,} bytes: >>>".format(
                resp.status_code, len(resp.content)
            )
        )

        for k, v in resp.headers.items():
            res.append("{}: {}".format(k, v))
        res.append("- " * 20)

        if short:
            s = shorten_string(s, 500, 100, place_holder="\n[...]\n")
        res.append(s)
        res.append("<<< " + "-" * 36)
        return "\n".join(res)

    @classmethod
    def _raise_assertion(cls, cause, resp):
        msg = cls._format_response(resp, short=True)
        msg = "\n  | ".join(msg.split("\n"))
        raise ActivityAssertionError("{}\n{}".format(cause, msg))

    def execute(self, session, **expanded_args):
        """
        Raises:
            ActivityAssertionError:
            requests.exceptions.ConnectionError: 'Connection refused', etc.
            requests.exceptions.HTTPError: On 404, 500, etc.
        """
        url = expanded_args.get("url")
        base_url = session.get_context("base_url")
        if not base_url and is_relative_url(url):
            raise ActivityError(
                "Missing context variable 'base_url' to resolve relative URLs"
            )
        expanded_args.setdefault("timeout", session.get_context("request_timeout"))
        assert "timeout" in expanded_args
        debug = expanded_args.get("debug")

        # print("session.dry_run", session.dry_run)
        if session.dry_run:
            return expanded_args.get("mock_result", "dummy_result")

        bs = session.browser_session
        method = self.raw_args["method"]

        url = expanded_args.pop("url")
        url = resolve_url(base_url, url)

        r_args = {k: v for k, v in expanded_args.items() if k in self.REQUEST_ARGS}

        verify_ssl = session.sessions.get("verify_ssl", True)
        r_args.setdefault("verify", verify_ssl)

        basic_auth = session.sessions.get("basic_auth", False)
        if basic_auth:
            r_args.setdefault("auth", session.user.auth)

        headers = r_args.setdefault("headers", {})
        headers.setdefault(
            "User-Agent",
            "session/{} Stressor/{}".format(session.session_id, __version__),
        )

        # if debug:
        #     http_client.HTTPConnection.debuglevel = 1
        # else:
        #     http_client.HTTPConnection.debuglevel = 0
        if debug:
            logger.info("HTTPRequest({}, {}, {})...".format(method, url, r_args))

        # The actual HTTP request:
        try:
            resp = bs.request(method, url, **r_args)
        except requests.exceptions.Timeout as e:
            raise ActivityTimeoutError("{}".format(e))
        except RequestException as e:
            raise ActivityError("{}".format(e))

        is_json = False
        try:
            result = resp.json()
            is_json = True
        except ValueError:
            result = resp.text

        if not resp.ok:
            logger.error(self._format_response(resp, short=not debug))
        elif debug:
            logger.info(self._format_response(resp, short=False))

        assert_status = expanded_args.get("assert_status")
        if not assert_status:
            # requests.exceptions.HTTPError: On 404, 500, etc.
            resp.raise_for_status()
        elif resp.status_code not in assert_status:
            self._raise_assertion(
                "HTTP status does not match {}: {}".format(
                    assert_status, resp.status_code
                ),
                resp,
            )

        arg = expanded_args.get("assert_match_headers")
        if arg:
            text = str(resp.headers)
            if not re.match(arg, text):
                self._raise_assertion(
                    "Result headers do not match `{}`".format(arg), resp
                )

        arg = expanded_args.get("assert_json")
        if arg:
            if not is_json:
                self._raise_assertion("Unexpected result type (expected JSON)", resp)

            for key, pattern in arg.items():
                value = get_dict_attr(result, key)
                # print(result, key, value)
                match, msg = match_value(pattern, value, key)
                if not match:
                    self._raise_assertion("Unexpected JSON result {}".format(msg), resp)

        arg = expanded_args.get("assert_html")
        if arg:
            if is_json:
                self._raise_assertion("Unexpected result type (expected HTML)", resp)

            for xpath, pattern in arg.items():
                # print(result)
                tree = html.fromstring(result)
                # print("tree", html.tostring(tree))
                # match = tree.xpath("body/span[contains(@class, 'test') and text() = 'abc']")
                match = tree.xpath(xpath)
                if pattern is True:
                    ok = bool(match)
                elif pattern is False:
                    ok = not match
                elif not match:
                    ok = False
                else:
                    # print("match", html.tostring(match))
                    raise NotImplementedError
                if not ok:
                    self._raise_assertion(
                        "Unexpected HTML result: XPath {!r} -> {}".format(xpath, match),
                        resp,
                    )

        return result


class GetRequestActivity(HTTPRequestActivity):
    def __init__(self, config_manager, **activity_args):
        super().__init__(config_manager, method="GET", **activity_args)


class PostRequestActivity(HTTPRequestActivity):
    def __init__(self, config_manager, **activity_args):
        super().__init__(config_manager, method="POST", **activity_args)


class PutRequestActivity(HTTPRequestActivity):
    def __init__(self, config_manager, **activity_args):
        super().__init__(config_manager, method="PUT", **activity_args)


class DeleteRequestActivity(HTTPRequestActivity):
    def __init__(self, config_manager, **activity_args):
        super().__init__(config_manager, method="DELETE", **activity_args)


class StaticRequestsActivity(ActivityBase):
    """
    This activity recieves a list of URLs (JavaScript, Html, CSS, Images, ...)
    and loads them, using max. ~5 threads, as a browser would.
    """

    REQUEST_ARGS = {"auth", "data", "json", "headers", "params", "timeout", "verify"}
    _mandatory_args = {"url_list"}
    _known_args = REQUEST_ARGS | _mandatory_args | {"assert_status", "thread_count"}
    _info_args = ("name", "method", "thread_count")

    def __init__(self, config_manager, **activity_args):
        super().__init__(config_manager, **activity_args)

    def execute(self, session, **expanded_args):
        """"""
        url_list = expanded_args.pop("url_list")
        base_url = session.get_context("base_url")
        expanded_args.setdefault("timeout", session.get_context("request_timeout"))
        debug = expanded_args.get("debug")
        thread_count = int(expanded_args.get("thread_count", 1))
        method = "GET"
        r_args = {k: v for k, v in expanded_args.items() if k in self.REQUEST_ARGS}

        verify_ssl = session.sessions.get("verify_ssl", True)
        r_args.setdefault("verify", verify_ssl)

        basic_auth = session.sessions.get("basic_auth", False)
        if basic_auth:
            r_args.setdefault("auth", session.user.auth)

        headers = r_args.setdefault("headers", {})
        headers.setdefault(
            "User-Agent",
            "session/{} Stressor/{}".format(session.session_id, __version__),
        )

        # TODO: requests.Session is not guaranteed to be thread-safe!
        bs = session.browser_session
        # Queue-up all pending request
        queue = Queue()
        for url in url_list:
            url = resolve_url(base_url, url)
            queue.put(url)

        results = []

        def _work(name):
            # logger.debug("StaticRequests({}) started...".format(name, ))
            while not session.stop_request.is_set():
                try:
                    url = queue.get(False)
                except Empty:
                    break

                if debug:
                    logger.info("StaticRequests({}, {})...".format(name, url))
                # The actual HTTP request:
                # TODO: requests.Session is not guaranteed to be thread-safe!
                try:
                    res = bs.request(method, url, **r_args)
                    res.raise_for_status()
                    results.append((True, name, url, None))
                except Exception as e:
                    results.append((False, name, url, "{}".format(e)))
                queue.task_done()
            logger.debug("StaticRequests({}) stopped.".format(name))
            return

        logger.debug(
            "Starting {} StaticRequestsActivity workers...".format(thread_count)
        )

        thread_list = []
        for i in range(thread_count):
            name = "{}.{:02}".format(session.session_id, i + 1)
            t = threading.Thread(name=name, target=_work, args=[name])
            t.setDaemon(True)  # Required to make Ctrl-C work
            thread_list.append(t)

        for t in thread_list:
            t.start()

        logger.debug("All StaticRequestsActivity workers running...")
        queue.join()
        for t in thread_list:
            t.join()
        errors = ["{}".format(error) for ok, name, url, error in results if not ok]
        if errors:
            raise ActivityError(
                "{} reqests failed:\n{}".format(len(errors), format(errors))
            )
            # logger.error(pformat(errors))
        return bool(errors)


class PollRequestActivity(HTTPRequestActivity):
    """
    TODO: This activity simulates a peridodical request to a single URL.
    """

    def __init__(self, config_manager, **activity_args):
        raise NotImplementedError
        # TODO: honor force_single option



# -*- coding: utf-8 -*-
# (c) 2020-2023 Martin Wendt and contributors; see https://github.com/mar10/stressor
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
import itertools
import sys
import threading
import time
from collections import defaultdict
from datetime import datetime

from snazzy import emoji, green, red, yellow

from stressor.config_manager import ConfigManager
from stressor.monitor.server import MonitorServer
from stressor.session_manager import SessionManager, User
from stressor.statistic_manager import StatisticManager
from stressor.util import (
    check_arg,
    format_elap,
    format_num,
    get_random_number,
    logger,
    set_console_ctrl_handler,
)


class RunManager:
    """
    Executes a run-configuration in parallel sessions.
    """

    CTRL_HANDLER_SET = None
    CURRENT_RUN_MANAGER = None

    DEFAULT_OPTS = {
        "monitor": False,
        "log_summary": True,
        # "dry_run": False,
    }
    STAGES = (
        # "new",
        "ready",
        "running",
        "done",
        "waiting",
        "stopping",
        "stopped",
    )
    CHANNELS = (
        "log",
        "start_run",
        "start_session",
        "start_sequence",
        "start_activity",
        "end_activity",
        "end_sequence",
        "end_session",
        "end_run",
    )
    activity_map = {}

    def __init__(self):
        self.lock = threading.RLock()
        self.host_id = "h1"
        self.process_id = "p1"
        #: :class:`ConfigManager` used to load the configuration YAML
        self.config_manager = None
        self.has_hooks = False
        self.has_catch_all_hooks = False
        self._hooks = defaultdict(list)
        #: Set this event to shut down the app
        self.stop_request = threading.Event()
        #: (bool): TODO: determines if a stop request is graceful or not
        #: True: Finalize the current sequence, then do 'end' sequence before stopping
        self.stop_request_graceful = None
        #: (bool): TODO: determines if a stop request will keep the monitor running
        #: True: Finalize the current sequence, then do 'end' sequence before stopping?
        self.stop_request_monitor = None
        self.session_list = []
        #: :class:`~stressor.statistic_manager.StatisticManager` object that containscurrent execution path
        self.stats = StatisticManager()
        self.options = self.DEFAULT_OPTS.copy()
        self.stage = "ready"
        self.start_dt = None
        self.start_stamp = None
        self.end_dt = None
        self.end_stamp = None

        # register_plugins()
        self.CURRENT_RUN_MANAGER = self
        self.set_console_ctrl_handler()

    def __str__(self):
        # name = self.config_manager.path if self.config_manager else "?"
        # name = self.run_config.get("name") if self.run_config else "?"
        return "RunManager<{}>".format(self.stage.upper())

    @staticmethod
    def set_console_ctrl_handler():
        if RunManager.CTRL_HANDLER_SET is None:
            RunManager.CTRL_HANDLER_SET = set_console_ctrl_handler(
                RunManager._console_ctrl_handler
            )
            logger.info("set_console_ctrl_handler()")

    @staticmethod
    def _console_ctrl_handler(ctrl):
        # NOTE: seems that print/logger do not work here?
        # print("_console_ctrl_handler()")
        return RunManager.CURRENT_RUN_MANAGER.console_ctrl_handler(ctrl)

    def console_ctrl_handler(self, ctrl):
        """
        Args:
            ctrl (int): 0: CTRL_C_EVENT, 1: CTRL_BREAK_EVENT, 2: CTRL_CLOSE_EVENT
        Returns:
            True if handled
            False if not handled, i.e. next registered handler will be called
        """
        # if self.stop_request.is_set():
        #     print("Got Ctrl-C 2nd time: terminating...")
        #     logger.warning("Got Ctrl-C a 2nd time: terminating...")
        #     time.sleep(0.1)
        #     # sys.exit(2)
        print("Got Ctrl-C (windows handler), terminating...", file=sys.stderr)
        logger.warning("Got Ctrl-C (windows handler), terminating...")
        # self.stop_request.set()
        self.stop()
        # return False
        return True

    def set_stage(self, stage):
        check_arg(stage, str, stage in self.STAGES)
        logger.info("Enter stage '{}'".format(stage.upper()))
        self.stage = stage

    def publish(self, channel, allow_cancel=False, *args, **kwargs):
        """Notify all subscribed handlers."""
        assert channel in self.CHANNELS
        result_list = []
        if not self.has_hooks:
            return result_list

        channel_hooks = self._hooks.get(channel)
        generic_hooks = self._hooks.get("*")
        if channel_hooks:
            if generic_hooks:
                hooks = itertools.chain(channel_hooks, generic_hooks)
            else:
                hooks = channel_hooks
        elif generic_hooks:
            hooks = generic_hooks
        for handler in hooks:
            res = handler(channel, *args, **kwargs)
            if allow_cancel and res is False:
                return False
            result_list.append(res)
        return result_list

    def subscribe(self, channel, handler):
        self.has_hooks = True
        if channel == "*":
            self.has_catch_all_hooks = True
        else:
            assert channel in (self.CHANNELS)
        assert callable(handler)
        self._hooks[channel].append(handler)

    def has_errors(self, or_warnings=False):
        return self.stats.has_errors()

    def get_cli_summary(self):
        cm = self.config_manager
        lines = []
        run_time = self.end_stamp - self.start_stamp
        user_count = len(self.session_list)
        has_errors = self.has_errors()

        ap = lines.append
        col = red if has_errors else green
        horz_line = col("=-" * 38 + "=")

        ap("Result Summary:")
        ap(horz_line)
        ap("Stressor scenario '{}' finished.".format(cm.name))
        ap("  Tag:      '{}'".format(cm.config.get("tag", "n.a.")))
        ap("  Base URL: {}".format(cm.config.get("base_url", "")))
        ap("  Start:    {}".format(self.start_dt.strftime("%Y-%m-%d %H:%M:%S")))
        ap("  End:      {}".format(self.end_dt.strftime("%Y-%m-%d %H:%M:%S")))
        ap(
            "Run time {}, net: {}.".format(
                format_elap(run_time, high_prec=True),
                format_elap(self.stats["net_act_time"], high_prec=True),
            )
        )
        ap(
            "Executed {:,} activities in {:,} sequences, using {:,} parallel sessions.".format(
                self.stats["act_count"],
                self.stats["seq_count"],
                user_count,
            )
        )
        if run_time and self.stats["seq_count"]:
            ap(
                "Sequence duration: {} average.".format(
                    format_elap(run_time / self.stats["seq_count"], high_prec=True)
                )
            )
        if run_time:
            ap(
                "             rate: {} sequences per minute (per user: {}).".format(
                    format_num(60.0 * self.stats["seq_count"] / run_time),
                    format_num(
                        60.0 * self.stats["seq_count"] / (run_time * user_count)
                    ),
                )
            )
            ap(
                "Activity rate:     {} activities per second (per user: {}).".format(
                    format_num(self.stats["act_count"] / run_time),
                    format_num(self.stats["act_count"] / (run_time * user_count)),
                )
            )

        # --- List of all activities that are marked `monitor: true`
        if self.stats["monitored"]:
            print(self.stats["monitored"])
            ap("{} monitored activities:".format(len(self.stats["monitored"])))
            for path, info in self.stats["monitored"].items():
                errors = info.get("errors")
                ap("  - {}".format(path))
                if not info:
                    ap("    n: 0, min: n.a., avg: n.a., max: n.a.")
                    continue

                ap(
                    "    n: {:,}, min: {}, avg: {}, max: {}{}".format(
                        info["act_count"],
                        format_elap(info["act_time_min"], high_prec=True),
                        format_elap(info["act_time_avg"], high_prec=True),
                        format_elap(info["act_time_max"], high_prec=True),
                        red(", {} errors".format(errors)) if errors else "",
                    )
                )

        if has_errors:
            pics = emoji(" 💥 💔 💥", "")
            ap(
                red(
                    "Result: ERROR, found {:,} errors and {:,} warnings.".format(
                        self.stats["errors"],
                        self.stats["warnings"],
                    )
                    + pics
                )
            )
            if self.stats.stats["run_limit_reached"]:
                ap(
                    yellow(
                        "Some activities where skipped due to the `max-errors` "
                        " or `max-time` limit."
                    )
                )
        else:
            pics = emoji(" ✨ 🍰 ✨", "")
            ap(green("Result: Ok." + pics))
        ap(horz_line)
        return "\n".join(lines)

    def get_status_info(self):
        cm = self.config_manager

        stats_info = self.stats.get_monitor_info(cm.config_all)

        res = {
            "name": cm.name,
            "scenarioDetails": cm.get("config.details", "n.a."),
            "tag": cm.get("config.tag", "n.a."),
            "stage": self.stage,
            "stageDisplay": "done" if self.stage == "waiting" else self.stage,
            "hasErrors": self.has_errors(),
            "startTimeStr": "{}".format(self.start_dt.strftime("%Y-%m-%d %H:%M:%S")),
            "baseUrl": cm.get("config.base_url"),
            "sessionCount": self.stats["sess_count"],
            "sessionsRunning": self.stats["sess_running"],
            "stats": stats_info,
        }
        if self.end_dt:
            elap = self.end_dt - self.start_dt
            res["endTimeStr"] = "{} ({})".format(
                self.end_dt.strftime("%Y-%m-%d %H:%M:%S"),
                format_elap(elap.total_seconds()),
            )
        else:
            elap = datetime.now() - self.start_dt
            res["endTimeStr"] = "(running for {}...)".format(
                format_elap(elap.total_seconds())
            )

        return res

    def log_info(self, *args, **kwargs):
        self.publish("log", level="info", *args, **kwargs)

    def load_config(self, run_config_file):
        """Load configuration file and set shortcuts."""
        cr = ConfigManager(self.stats)
        cr.read(run_config_file, load_files=True)

        self.config_manager = cr
        # self.run_config = cr.run_config
        logger.info("Successfully compiled configuration {}.".format(cr.path))

    def _run_one(self, session_manager):
        """Run inside a separate thread."""
        try:
            session_manager.run()
            # We don't need to print results if only one session was run, since
            # it is also part of the global stats:
            # rc = self.run_config
            # if not rc.get("force_single") and rc["sessions"]["count"] > 1:
            #     logger.info(
            #         "Results for {}:\n{}".format(
            #             session_manager, session_manager.stats.format_result()
            #         )
            #     )
        except KeyboardInterrupt:
            logger.exception("Session thread received Ctrl-C")
            self.stats.report_error(None, None, None, "KeyboardInterrupt")
            self.stop_request.set()
            # self.stats.inc("errors")
        except Exception as e:
            logger.exception("Session thread raised exception")
            self.stats.report_error(None, None, None, e)
            # self.stats.inc("errors")
            # raise
        return

    def run_in_threads(self, user_list, context):
        self.publish("start_run", run_manager=self)
        self.stop_request.clear()
        thread_list = []
        self.session_list = []
        for i, user in enumerate(user_list, 1):
            name = "t{:02}".format(i)
            sess = SessionManager(self, context, name, user)
            self.session_list.append(sess)
            t = threading.Thread(name=name, target=self._run_one, args=[sess])
            t.setDaemon(True)  # Required to make Ctrl-C work
            thread_list.append(t)

        logger.info("Starting {} session workers...".format(len(thread_list)))
        self.set_stage("running")
        self.stats.report_start(None, None, None)

        ramp_up_delay = self.config_manager.sessions.get("ramp_up_delay")

        start_run = time.monotonic()
        for i, t in enumerate(thread_list):
            if ramp_up_delay and i > 1:
                delay = get_random_number(ramp_up_delay)
                logger.info(
                    "Ramp-up delay for t{:02}: {:.2f} seconds...".format(i, delay)
                )
                time.sleep(delay)
            t.start()

        logger.important(
            "All {} sessions running, waiting for them to terminate...".format(
                len(thread_list)
            )
        )
        for t in thread_list:
            t.join()

        self.set_stage("done")
        elap = time.monotonic() - start_run

        # self.stats.add_timing("run", elap)
        self.stats.report_end(None, None, None)

        self.publish("end_run", run_manager=self, elap=elap)

        logger.debug("Results for {}:\n{}".format(self, self.stats.format_result()))

        return not self.has_errors()

    def run(self, options, extra_context=None):
        """Run the current

        Args:
            options (dict): see RunManager.DEFAULT_OPTS
            extra_context (dict, optional):
        Returns:
            (int) Exit code 0 if no errors occurred
        """
        check_arg(options, dict)
        check_arg(extra_context, dict, or_none=True)

        self.options.update(options)

        if extra_context:
            self.config_manager.update_config(extra_context)

        context = self.config_manager.context
        sessions = self.config_manager.sessions

        count = int(sessions.get("count", 1))
        if count > 1 and self.config_manager.config.get("force_single"):
            logger.info("force_single: restricting sessions count to one.")
            count = 1

        # Construct a `User` with at least 'name', 'password', and optional
        # custom attributes
        user_list = []
        for user_dict in sessions["users"]:
            user = User(**user_dict)
            user_list.append(user)
        # We have N users and want `count` sessions: re-use round-robin
        user_list = itertools.islice(itertools.cycle(user_list), 0, count)
        user_list = list(user_list)

        monitor = None
        if self.options.get("monitor"):
            monitor = MonitorServer(self)
            monitor.start()
            time.sleep(0.5)
            monitor.open_browser()

        self.start_stamp = time.monotonic()
        self.start_dt = datetime.now()
        self.end_dt = None
        self.end_stamp = None
        try:
            try:
                res = False
                res = self.run_in_threads(user_list, context)
            except KeyboardInterrupt:
                # if not self.stop_request.is_set():
                logger.warning("Caught Ctrl-C: terminating...")
                self.stop()
            finally:
                self.end_dt = datetime.now()
                self.end_stamp = time.monotonic()

            if self.options.get("log_summary", True):
                logger.important(self.get_cli_summary())

            if monitor:
                self.set_stage("waiting")
                logger.important("Waiting for monitor... Press Ctrl+C to quit.")
                self.stop_request.wait()
        finally:
            if monitor:
                monitor.shutdown()

            # print("RES", res, self.has_errors(), self.stats.format_result())
            self.set_stage("stopped")
        return res

    def stop(self, graceful=2):
        """"""
        # logger.info("Stop request received")
        # TODO: set errors += 1 if we interrupt a running stage
        self.set_stage("stopping")
        self.stop_request.set()
        return True

    def get_run_time(self):
        return time.monotonic() - self.start_stamp



# -*- coding: utf-8 -*-
# (c) 2020-2023 Martin Wendt and contributors; see https://github.com/mar10/stressor
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
import io
import json
import logging
import os
import socketserver
import webbrowser
from http.server import HTTPStatus, SimpleHTTPRequestHandler
from threading import Thread
from urllib import parse

from stressor import __version__

# from stressor.util import logger


logger = logging.getLogger("stressor.monitor")


class Handler(SimpleHTTPRequestHandler):
    server_version = (
        "stressor/" + __version__ + " " + SimpleHTTPRequestHandler.server_version
    )
    # Custom attributes, set by `MonitorServer`:
    DIRECTORY = None
    run_manager = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, directory=self.DIRECTORY, **kwargs)

    def log_request(self, code="-", size="-"):
        # Overide base implementation (writing to stderr)
        if isinstance(code, HTTPStatus) and not (200 <= code.value < 400):
            logger.warning('"%s" %s %s', self.requestline, str(code), str(size))
        else:
            logger.debug('"%s" %s %s', self.requestline, str(code), str(size))

    def log_error(self, format, *args):
        # Overide base implementation (writing to stderr)
        logger.error(format, *args)

    def log_message(self, format, *args):
        # Overide base implementation (writing to stderr)
        logger.debug(format, *args)

    def _return_json(self, body, status=HTTPStatus.OK):
        if not isinstance(body, str):
            body = json.dumps(body)
        encoded = body.encode("utf8", "surrogateescape")
        f = io.BytesIO()
        f.write(encoded)
        f.seek(0)
        # print(encoded)
        self.send_response(status)
        self.send_header("Content-type", "application/json; charset=utf-8")
        self.send_header("Content-Length", str(len(encoded)))
        self.end_headers()
        try:
            self.copyfile(f, self.wfile)
        finally:
            f.close()
        return f

    def on_stopManager(self, args):
        res = self.run_manager.stop()
        return self._return_json(res)

    def on_getStats(self, args):
        res = self.run_manager.get_status_info()
        return self._return_json(res)

    def on_getErrorInfo(self, args):
        res = self.run_manager.stats.get_error_info(args)
        return self._return_json(res)

    def do_GET(self):
        handler_name = self.path.strip("/")
        handler_name, _sep, _args = handler_name.partition("?")
        args = dict(parse.parse_qsl(parse.urlsplit(self.path).query))
        handler = getattr(self, "on_" + handler_name, None)
        if callable(handler):
            try:
                return handler(args)
            except Exception as e:
                logger.exception(handler_name)
                return self._return_json(
                    {"fault": repr(e)}, HTTPStatus.INTERNAL_SERVER_ERROR
                )

        return SimpleHTTPRequestHandler.do_GET(self)


class MonitorServer(Thread):
    """
    Run a web server in a separate thread, so it does not block
    """

    def __init__(self, run_manager, bind="", port=8081):
        super().__init__(name="stressor.monitor", daemon=None)
        Handler.DIRECTORY = os.path.join(os.path.dirname(__file__), "htdocs")
        Handler.run_manager = run_manager
        self.run_manager = run_manager
        self.bind = bind
        self.port = port
        self.httpd = None

    def run(self):
        with socketserver.TCPServer((self.bind, self.port), Handler) as httpd:
            self.httpd = httpd
            logger.info(
                "Monitor serving at http://{}:{}...".format(
                    self.bind or "localhost", self.port
                )
            )
            httpd.serve_forever()
        self.httpd = None
        logger.info("Monitor server stopped.")

    def shutdown(self):
        if self.httpd:
            self.httpd.shutdown()

    def open_browser(self):
        assert self.bind == ""
        monitor_url = "http://localhost:{}/".format(self.port)
        # monitor_url = "http://127.0.0.1:{}/".format(self.port)
        logger.info("Open web browser at {}".format(monitor_url))
        webbrowser.open_new_tab(monitor_url)



# -*- coding: utf-8 -*-
# (c) 2020-2023 Martin Wendt and contributors; see https://github.com/mar10/yabs
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
from pkg_resources import iter_entry_points

from stressor.plugins.base import ActivityBase, MacroBase
from stressor.util import logger


class PluginManager:
    """
    Load, cache, and maintain a list of plugins and workflow tasks.

    This is basically a singleton.
    """

    #: (dict) Currently known activity classes by their script name,
    #: e.g. {'GetRequest': GetRequestActivity, ...}
    #: (Call :func:`register_activity_plugins()` to update this map after new plugins
    #: have been imported.)
    activity_plugin_map = {}

    #: (dict) Currently known macro classes by their script name,
    #: e.g. {'load': LoadMacro, ...}
    #: (Call :func:`register_macro_plugins()` to update this map after new plugins
    #: have been imported.)
    macro_plugin_map = {}

    #: (str) Entry point group name
    namespace = "stressor.plugins"
    #: (bool) True when all plugin entry points are loaded
    entry_points_searched = False
    #: (bool) True when all plugins' 'register()' method were called.
    plugins_registered = False
    #: (dict) Cached map TASK_NAME => EntryPoint of loaded entry_points
    _entry_point_map = {}

    def __init__(self):
        pass

    @classmethod
    def find_plugins(cls):
        """Load all entry points with group name 'stressor.plugins'."""
        if cls.entry_points_searched:
            return
        cls.entry_points_searched = True
        ep_map = cls._entry_point_map
        logger.debug("Search entry points for group '{}'...".format(cls.namespace))

        for ep in iter_entry_points(group=cls.namespace, name=None):
            plugin_name = "{}".format(ep.dist)
            logger.debug(
                "Found plugin {} from entry point `{}`".format(plugin_name, ep)
            )

            if ep.name in ep_map:
                logger.warning(
                    "Duplicate entry point name: {}; skipping...".format(ep.name)
                )
                continue
            # elif ep.name in cls.task_class_map:
            #     # TODO: support overriding standard tasks?
            #     # Maybe when 'extras=[override]' is passed...
            #     logger.warning(
            #         "Plugin task name already exists: {}; skipping...".format(ep.name)
            #     )
            #     continue

            ep_map[ep.name] = ep

        return

    @classmethod
    def _register_subclasses(cls, base_cls, cls_map):
        """Register currently loaded plugins.

        These are all currently known subclasses of
        :class:`stressor.plugins.base.ActivityBase` and
        :class:`stressor.plugins.base.MacroBase` .
        """
        for sub_cls in base_cls.__subclasses__():
            name = sub_cls.get_script_name()
            existing_cls = cls_map.get(name)
            if existing_cls is sub_cls:
                continue  # already registered
            elif existing_cls is None:
                if not name.startswith("_"):
                    cls_map[name] = sub_cls
                cls._register_subclasses(sub_cls, cls_map)
            else:
                raise RuntimeError("Class name conflict: {}".format(sub_cls))
        return

    @classmethod
    def register_plugins(cls, arg_parser):
        """Call `register_fn` on all loaded entry points."""
        if cls.plugins_registered:
            return
        cls.plugins_registered = True

        # Import stock class definitions, so we can scan the subclasses.
        # This will add standard plugins to ActivityBase.__subclasses__()
        # and MacroBase.__subclasses__():
        import stressor.plugins.common  # noqa F401
        import stressor.plugins.http_activities  # noqa F401
        import stressor.plugins.script_activities  # noqa F401

        # Load entry points from all installed mosules that have the
        # 'stressor.plugins' namespace:
        cls.find_plugins()

        # Call `register_fn` on all loaded entry points_
        ep_map = cls._entry_point_map
        for name, ep in cls._entry_point_map.items():
            logger.info("Load plugins {}...".format(ep.dist))
            try:
                register_fn = ep.load()
                if not callable(register_fn):
                    raise RuntimeError("Entry point {} is not a function".format(ep))
                ep_map[ep.name] = register_fn
            except Exception:
                logger.exception("Failed to load {}".format(ep))

            prev_activities = list(ActivityBase.__subclasses__())
            prev_macros = list(MacroBase.__subclasses__())

            logger.debug("Register plugins {}...".format(ep.dist))
            try:
                # The plugin must declare new classes derrived from
                # ActivityBase and/or MacroBase
                register_fn(
                    activity_base=ActivityBase,
                    macro_base=MacroBase,
                    arg_parser=arg_parser,
                )
            except Exception:
                logger.exception("Could not register {}".format(name))
                continue

            found_one = False
            for activity_cls in ActivityBase.__subclasses__():
                if activity_cls in prev_activities:
                    continue
                found_one = True
                logger.info("Register {}.{}".format(ep.dist, activity_cls))

            for macro_cls in MacroBase.__subclasses__():
                if macro_cls in prev_macros:
                    continue
                found_one = True
                logger.info("Register {}.{}".format(ep.dist, macro_cls))

            if not found_one:
                logger.warning(
                    "Plugin {} did not register activites nor macros".format(ep.dist)
                )

        # Build plugin maps from currently known subclasses
        cls._register_subclasses(ActivityBase, cls.activity_plugin_map)
        logger.debug("Registered activity plugins:\n{}".format(cls.activity_plugin_map))

        cls._register_subclasses(MacroBase, cls.macro_plugin_map)
        logger.debug("Registered macro plugins:\n{}".format(cls.macro_plugin_map))
        return



# -*- coding: utf-8 -*-
# (c) 2020-2023 Martin Wendt and contributors; see https://github.com/mar10/stressor
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
import os
from textwrap import dedent


class TestScripts:
    def setup_method(self):
        self.fixtures_path = os.path.join(os.path.dirname(__file__), "fixtures")

    def test_script(self):
        # config_path = os.path.join(self.fixtures_path, "test_mock_server.yaml")
        # rm = RunManager()
        # rm.load_config(config_path)
        # assert rm.run_config
        # assert rm.config_manager.file_version == 0
        # res = rm.run()
        # assert res is True
        script = """
        print("SCRIPT: globals()", globals().keys())
        # print("SCRIPT: globals()", globals()["foo"])
        # print("SCRIPT:", globals()["foo"])
        # # print("Hello world!", locals()[42])
        # print("Hello world!", local)
        # print("Script globsals()!", globals())
        b = 18
        my_global += "script"
        my_local += "script"
        17
        result = 42
        """

        # TODO: Compile?

        # local_vars = locals()
        global_vars = {
            "my_global": "org",
            # "__builtins__": {},
        }
        local_vars = {
            "my_local": "org",
        }
        prev_builtins = set(__builtins__.keys())
        res = exec(
            dedent(script),
            global_vars,
            # {
            #     # "__builtins__": None,
            #     "foo": "bar",
            # },
            local_vars,
        )
        assert res is None
        print("New built-ins:", set(__builtins__.keys()).difference(prev_builtins))
        print("Result: {}".format(res))
        print("Globals: {}".format(global_vars.keys()))
        print("Locals: {}".format(local_vars))
        # assert 0



# -*- coding: utf-8 -*-
# (c) 2020-2023 Martin Wendt and contributors; see https://github.com/mar10/stressor
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
import os
import random
from textwrap import dedent

import yaml

from stressor.plugins.base import ActivityBase, MacroBase
from stressor.util import check_arg, format_elap


class LoadMacro(MacroBase):
    """Implement `$load(path)` macro."""

    def apply(self, config_manager, parent, parent_key, path):
        path = config_manager.resolve_path(path, must_exist=True)

        if path.lower().endswith(".py"):
            assert parent_key == "script"
            with open(path, "rt") as f:
                res = f.read()
            parent[parent_key] = res
            return res

        if not path.lower().endswith((".yaml", ".yml")):
            raise NotImplementedError
        # Load (and )
        with open(path, "rt") as f:
            res = yaml.safe_load(f)
        assert isinstance(parent, dict)
        assert isinstance(res, list)
        parent[parent_key] = res
        # print(yaml.dump(parent, default_flow_style=False))
        return res


class EnvMacro(MacroBase):
    """Implement `$env(var_name)` macro, which resolves an environment variable at load-time."""

    def apply(self, config_manager, parent, parent_key, var_name):
        # TODO:
        # Allow optional 'default' args: `$env(HOME, Foo)`
        # Maybe introspection can automate this?
        # var_name, default = parse_arglist(var_name)
        res = os.environ[var_name]
        parent[parent_key] = res


class DebugMacro(MacroBase):
    """Implement `$debug()` macro, which dumps information at run-time."""

    def apply(self, config_manager, parent, parent_key, var_name):
        parent[parent_key] = "RunScript"
        parent["name"] = "$debug"
        parent["export"] = []
        parent["script"] = dedent(
            r"""
            from pprint import pprint
            pprint(locals())
        """
        )


# class CrontabMacro(MacroBase):
#     """Implement `$cron()` macro, which inserts the current UTC time stamp.

#     https://github.com/josiahcarlson/parse-crontab
#     """

#     def apply(self, config_manager, parent, parent_key, var_name):
#         raise NotImplementedError


# class StampMacro(MacroBase):
#     """Implement `$stamp()` macro, which inserts the current UTC time stamp."""

#     run_time_eval = True

#     def apply(self, config_manager, parent, parent_key, var_name):
#         raise NotImplementedError


# class StopMacro(MacroBase):
#     """Implement `$stop()` macro, which dumps information at run-time and ends execution."""

#     def apply(self, config_manager, parent, parent_key, var_name):
#         raise NotImplementedError


# class BreakMacro(MacroBase):
#     """Implement `$break()` macro, which halts and enters single-step mode."""

#     def apply(self, config_manager, parent, parent_key, var_name):
#         # dump info
#         # enter single-step mode
#         # allow to stop the run
#         raise NotImplementedError


class SleepMacro(MacroBase):
    """Implement `$sleep(duration)` macro, which is a shortcut to :class`SleepActivity`."""

    _args_def = (
        ("min", float),  # mandatory
        ("max", float, None),  # optional
    )

    def apply(self, config_manager, parent, parent_key, min, max):
        # duration is always a  string, it may even be $(context.var) macro.
        # We still do some checking here, to get early load-time errors.
        # check_arg(min, str)

        assert parent_key == "activity"
        # if "$" not in min:
        #     assert_always(float(min) >= 0)

        parent[parent_key] = "Sleep"
        parent["duration"] = min
        if max is not None:
            parent["duration_2"] = max

        return True


class SleepActivity(ActivityBase):
    """
    Args:
        duration (float): sleep duration in seconds

    Examples::

        sequences:
          main:
            - activity: Sleep
              duration: 0.5
    """

    _mandatory_args = {"duration"}
    _known_args = {"duration", "duration_2"}
    _info_args = ("name", "duration", "duration_2")

    #: Sleep activities are ignorable by default
    _default_ignore_timing = True

    def __init__(self, config_manager, **activity_args):
        check_arg(activity_args.get("duration"), (str, int, float))
        check_arg(activity_args.get("duration_2"), (str, int, float), or_none=True)

        super().__init__(config_manager, **activity_args)

        # TODO: Support CronTab syntax
        #     https://github.com/taichino/croniter
        #     https://github.com/josiahcarlson/parse-crontab

    # def __str__(self):
    #     return "Sleep({:.3} sec.)".format(self.duration)

    def get_info(self, info_args=True, expanded_args=None, session=None):
        # We want to display the actual sleep duration even if it was calculated
        # from a range. So we store it in `prepare_execute` and use it here
        # if available.
        if (
            expanded_args and session and "duration_2" in expanded_args
        ):  # Set by prepare_execute()
            return "{}(duration[{}..{}] => {})".format(
                self.get_script_name(),
                float(expanded_args["duration"]),
                float(expanded_args["duration_2"]),
                format_elap(session.data["_cur_duration"]),
            )
        return super().get_info(info_args, expanded_args, session)

    def prepare_execute(self, session, expanded_args):
        duration = float(expanded_args["duration"])
        duration_2 = expanded_args.get("duration_2")
        if duration_2 is not None:
            duration = random.uniform(duration, float(duration_2))
        session.data["_cur_duration"] = duration

    def execute(self, session, **expanded_args):
        assert "_cur_duration" in session.data
        duration = session.data["_cur_duration"]
        session.data["_cur_duration"] = None
        if not session.dry_run:
            session.stop_request.wait(timeout=duration)
        return


# class MemoryActivity(ActivityBase):
#     """
#     Args:
#         operation (str): 'alloc' | 'free'
#         size (int):
#     """

#     def __init__(self, config_manager, operation, **activity_args):
#         super().__init__(config_manager, **activity_args)

#     def execute(self, session):
#         raise NotImplementedError


# class CpuActivity(ActivityBase):
#     """
#     Args:
#         operation (str): 'load' | 'stop'
#         cores (int): Default: 1
#         async (bool): Default: True
#         duration (float):
#     """

#     def __init__(self, config_manager, operation, **activity_args):
#         super().__init__(config_manager, **activity_args)

#     def execute(self, session):
#         raise NotImplementedError


# class FileActivity(ActivityBase):
#     """
#     Args:
#         operation (str): 'read' | 'write'
#         size (int):
#         path (str):
#     """

#     def __init__(self, config_manager, operation, **activity_args):
#         super().__init__(config_manager, **activity_args)

#     def execute(self, session):
#         raise NotImplementedError



# -*- coding: utf-8 -*-
"""
Stress-test your web app.

(c) 2020-2023 Martin Wendt and contributors; see https://github.com/mar10/stressor
Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php

Usage examples:
    $ stressor --help
    $ stressor run .
"""
import argparse
import logging
import os
import sys

import yaml
from snazzy import enable_colors

from stressor import __version__
from stressor.cli_common import common_parser, verbose_parser
from stressor.convert.har_converter import HarConverter
from stressor.plugin_manager import PluginManager
from stressor.run_manager import RunManager
from stressor.util import (
    check_cli_verbose,
    init_logging,
    logger,
    parse_option_args,
    version_info,
)


def handle_run_command(parser, args):
    options = {
        "monitor": args.monitor,
        "log_summary": True,
        # "dry_run": args.dry_run,
    }
    extra_context = {}
    # Parse `--option=NAME:VALUE` arguments:
    try:
        extra_context.update(parse_option_args(args.option, coerce_values=True))
    except Exception as e:
        parser.error("--option: {}".format(e))
    # Make sure that --dry-run arg is always honored:
    if args.dry_run:
        extra_context["dry_run"] = True

    scenario_fspec = args.scenario
    if os.path.isdir(scenario_fspec):
        scenario_fspec = os.path.join(scenario_fspec, "scenario.yaml")
        logger.info("Looking for {}".format(scenario_fspec))

    rm = RunManager()
    rm.load_config(scenario_fspec)
    if args.single:
        rm.config_manager.config["force_single"] = True
    if args.max_time:
        rm.config_manager.config["max_time"] = float(args.max_time)
    if args.max_errors:
        rm.config_manager.config["max_errors"] = int(args.max_errors)

    res = rm.run(options, extra_context)

    if not res:
        logger.error("Finished with errors.")
        return 1
    logger.info("Stressor run succesfully completed.")
    return 0


def handle_init_command(parser, args):
    opts = {}
    if args.opts:
        if not os.path.isfile(args.opts):
            parser.error("File not found: {}".format(args.opts))
        with open(args.opts, "rt") as f:
            opts = yaml.safe_load(f)

    opts.update(
        {
            "fspec": args.har_file,
            "target_folder": args.target,
            "force": args.force,
            "dry_run": args.dry_run,
        }
    )
    conv = HarConverter(opts)
    res = conv.run()
    return res


def handle_listen_command(parser, args):
    raise NotImplementedError


# ===============================================================================
# run
# ===============================================================================
def run():
    """CLI main entry point."""

    parser = argparse.ArgumentParser(
        description="Stress-test your web app.",
        epilog="See also https://github.com/mar10/stressor",
        parents=[verbose_parser, common_parser],
        # allow_abbrev=False,
    )
    parser.add_argument(
        "-V",
        "--version",
        action="store_true",
        help="display version info and exit (combine with -v for more information)",
    )
    subparsers = parser.add_subparsers(help="sub-command help")

    # --- Create the parser for the "run" command ------------------------------

    sp = subparsers.add_parser(
        "run",
        parents=[verbose_parser, common_parser],
        help="run a test suite scenario",
    )

    sp.add_argument(
        "scenario",
        metavar="SCENARIO",
        default="./scenario.yaml",
        help="path to configuration file or folder (default: %(default)s)",
    )
    sp.add_argument(
        "-o",
        "--option",
        action="append",
        help="override configuration, syntax `OPTION:VALUE` (multiple values allowed)",
    )
    sp.add_argument(
        "--single",
        action="store_true",
        help="Force `sessions.count: 1` and prevent sequence loops, so only one thread is run straight",
    )
    sp.add_argument(
        "--monitor",
        action="store_true",
        help="Open a web server and browser application to display real-time progress",
    )
    sp.add_argument(
        "--max-errors",
        type=int,
        default=0,
        help="Stop after N errors (overrides `config.max_errors`)",
    )
    sp.add_argument(
        "--max-time",
        type=float,
        default=0.0,
        help="Stop after N seconds (overrides `config.max_time`)",
    )

    sp.set_defaults(command=handle_run_command)

    # --- Create the parser for the "init" command ---------------------------

    sp = subparsers.add_parser(
        "init",
        parents=[verbose_parser, common_parser],
        help="create new scenario folder and optinally convert HAR files",
    )
    sp.add_argument(
        "target",
        metavar="TARGET",
        # default="./scenario.yaml",
        help="target folder (created if not existing)",
    )
    sp.add_argument(
        "--import",
        dest="har_file",
        # required=True,
        help="optional HAR file that is converted",
    )
    sp.add_argument(
        "--force",
        action="store_true",
        help="override existing files",
    )
    sp.add_argument(
        "--opts",
        help="YAML file with conversion options",
    )
    sp.set_defaults(command=handle_init_command)

    # --- Create the parser for the "listen" command ---------------------------

    # sp = subparsers.add_parser(
    #     "listen",
    #     parents=[verbose_parser, common_parser],
    #     help="run in 'drone' mode, listening for commands from master",
    # )

    # sp.add_argument(
    #     "--host",
    #     default="0.0.0.0",
    #     help="local ip address or hostname to bind to (default: %(default)s)",
    # )
    # sp.add_argument(
    #     "--port",
    #     type=int,
    #     default=8082,
    #     help="local port number to bind to (default: %(default)s)",
    # )
    # sp.add_argument(
    #     "--secret",
    #     default=None,
    #     help="password that master must use (default: random)",
    # )
    # sp.add_argument(
    #     "--collect",
    #     metavar="SECS",
    #     type=int,
    #     default=0,
    #     help="report ram, cpu load, and other system metrics every SECS seconds (default: off)",
    # )

    # sp.set_defaults(command=handle_listen_command)

    # --- Let all sublasses of `WorkflowTask` add their arguments --------------
    # We want to see some logging, even if init_logging() wasn't called yet:
    level = logging.DEBUG if check_cli_verbose() >= 4 else logging.INFO
    logging.basicConfig(level=level, format="%(message)s", datefmt="%H:%M:%S")

    PluginManager.register_plugins(parser)

    # --- Parse command line ---------------------------------------------------

    args = parser.parse_args()

    args.verbose -= args.quiet
    del args.quiet

    # print("verbose", args.verbose)
    init_logging(args.verbose, args.log_file)

    if not args.no_color:
        # Enable snazzy colors and emojis if terminal supports them
        enable_colors(True, force=False)

    if getattr(args, "version", None):
        if args.verbose >= 4:
            info = version_info
            info += "\nPython from: {}".format(sys.executable)
        else:
            info = __version__
        print(info)
        sys.exit(0)

    if not callable(getattr(args, "command", None)):
        parser.error("missing command")

    try:
        return args.command(parser, args)
    except KeyboardInterrupt:
        print("\nAborted by user.", file=sys.stderr)
        sys.exit(3)
    # Unreachable...
    return


# Script entry point
if __name__ == "__main__":
    # Just in case...
    from multiprocessing import freeze_support

    freeze_support()

    run()



# -*- coding: utf-8 -*-
# (c) 2020-2023 Martin Wendt and contributors; see https://github.com/mar10/stressor
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""

import os

import pytest

from stressor.context_stack import ContextStack


class TestContextStack:
    def setup_method(self):
        self.fixtures_path = os.path.join(os.path.dirname(__file__), "fixtures")
        self.context_stack = ContextStack()

    def teardown_method(self):
        pass

    def test_1(self):
        cm = self.context_stack
        with pytest.raises(IndexError):
            cm.peek()

        assert cm.path() == "/"

        cm.push("t1", {"root": "http://example.com", "url": "page_1"}, copy_data=True)
        assert cm.peek().name == "t1"
        assert cm.path() == "/t1"
        assert cm.as_dict() == {
            "root": "http://example.com",
            "url": "page_1",
        }

        cm.push("t2", {"url": "page_2"}, copy_data=True)
        assert cm.path() == "/t1/t2"
        assert cm.as_dict() == {
            "root": "http://example.com",
            "url": "page_2",
        }
        cm.set_last_part("t2_new")
        assert cm.path() == "/t1/t2_new"
        cm.set_last_part("t2")

        assert cm.peek().name == "t2"
        assert cm.pop().name == "t2"
        assert cm.peek().name == "t1"
        assert cm.path() == "/t1"
        assert cm.as_dict() == {
            "root": "http://example.com",
            "url": "page_1",
        }

        assert cm.pop().name == "t1"
        with pytest.raises(IndexError):
            cm.pop()
        with pytest.raises(IndexError):
            cm.peek()

        with pytest.raises(RuntimeError):
            for i in range(ContextStack.MAX_DEPTH + 1):
                cm.push("t{:02}".format(i), {})



# -*- coding: utf-8 -*-
# (c) 2020-2023 Martin Wendt and contributors; see https://github.com/mar10/stressor
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
import pytest

from stressor.util import (
    PathStack,
    assert_always,
    check_arg,
    coerce_str,
    format_elap,
    format_num,
    format_rate,
    get_dict_attr,
    get_random_number,
    is_yaml_keyword,
    parse_args_from_str,
    parse_option_args,
    shorten_string,
)


class TestBasics:
    def test_assert(self):
        assert_always(1)
        assert_always(1, "test")
        with pytest.raises(AssertionError):
            assert_always(0)
        with pytest.raises(AssertionError, match=".*foobar.*"):
            assert_always(0, "foobar")

    def test_check_arg(self):
        def foo(name, amount, options=None):
            check_arg(name, str)
            check_arg(amount, (int, float), amount > 0)
            check_arg(options, dict, or_none=True)
            return name

        assert foo("x", 10) == "x"
        assert foo("x", 10, {1: 2}) == "x"
        with pytest.raises(TypeError, match=".*but got <class 'int'>.*"):
            foo("x", 10, 42)
        with pytest.raises(ValueError, match=".*Invalid argument value.*"):
            foo("x", -10)
        return

    def test_get_random_number(self):
        assert get_random_number(None) is None
        assert get_random_number(1) == 1.0
        assert 0.1 <= get_random_number((0.1, 2.0)) <= 2.0
        with pytest.raises(TypeError):
            get_random_number("1")

    def test_is_yaml_keyword(self):
        assert is_yaml_keyword(None) is False
        assert is_yaml_keyword("") is False
        assert is_yaml_keyword("0") is False
        assert is_yaml_keyword("1") is False
        #: NOTE: YAML evaluates `0_` as "0" and `0_1_` as "1", so we don't accept leading numbers
        assert is_yaml_keyword("0_") is False
        assert is_yaml_keyword("_0") is True

    def test_get_attr(self):

        # next_
        class TestClass:  # noqa: B903
            def __init__(self, val):
                self.val = val

        d = {
            "s1": 1,
            "s2": "foo",
            "l1": [],
            "l2": [1, 2, 3],
            "d1": {},
            "d2": {"d21": 42, "d22": {"d221": "bar"}},
            "o": TestClass("baz"),
        }
        assert get_dict_attr(d, "s1") == 1
        assert get_dict_attr(d, "l1") == []
        assert get_dict_attr(d, "l2") == [1, 2, 3]
        assert get_dict_attr(d, "l2.[1]") == 2
        assert get_dict_attr(d, "d1") == {}
        assert get_dict_attr(d, "d2.d21") == 42
        assert get_dict_attr(d, "d2.d22.d221") == "bar"
        assert get_dict_attr(d, "o.val") == "baz"

        with pytest.raises(KeyError):
            get_dict_attr(d, "foobar")
        with pytest.raises(
            AttributeError, match=r".*object has no attribute 'foobar'.*"
        ):
            get_dict_attr(d, "s1.foobar")
        with pytest.raises(ValueError, match=r".*Use `\[INT\]` syntax.*"):
            get_dict_attr(d, "l2.foobar")
        with pytest.raises(IndexError):
            get_dict_attr(d, "l2.[99]")
        with pytest.raises(KeyError):
            get_dict_attr(d, "d1.foobar")

        # Defaults:
        assert get_dict_attr(d, "foobar", "def") == "def"
        assert get_dict_attr(d, "s1.foobar", "def") == "def"
        assert get_dict_attr(d, "l2.foobar", "def") == "def"
        assert get_dict_attr(d, "l2.[99]", "def") == "def"
        assert get_dict_attr(d, "foobar", "def") == "def"
        assert get_dict_attr(d, "d1.foobar", "def") == "def"

    def test_pathstack(self):
        path = PathStack()
        assert str(path) == "/"
        path.push("root")
        assert str(path) == "/root"
        with path.enter("sub1"):
            assert str(path) == "/root/sub1"
        assert str(path) == "/root"
        assert path.pop() == "root"
        with pytest.raises(IndexError):
            path.pop()

    def test_shorten_string(self):
        s = (
            "Do you see any Teletubbies in here?"
            "Do you see a slender plastic tag clipped to my shirt with my name printed on it?"
        )
        assert shorten_string(None, 10) is None
        assert len(shorten_string(s, 20)) == 20
        assert len(shorten_string(s, 20, min_tail_chars=7)) == 20
        assert shorten_string(s, 20) == "Do you see any [...]"
        assert shorten_string(s, 20, min_tail_chars=7) == "Do you s[...] on it?"

    def test_format_elap(self):
        assert format_elap(1.23456) == "1.23 sec"
        assert format_elap(1.23456, high_prec=True) == "1.235 sec"
        assert format_elap(3677) == "1:01:17 hrs"
        assert format_elap(367) == "6:07 min"
        assert format_elap(367, high_prec=True) == "6:07.00 min"
        assert format_elap(12.34, count=10) == "12.3 sec, 0.8 items/sec"

    def test_format_num(self):
        assert format_num(1000.2345) == "1,000"
        assert format_num(100.2345) == "100.2"
        assert format_num(10.2345) == "10.23"
        assert format_num(1.2345) == "1.234"
        assert format_num(0) == "0"
        assert format_num(1) == "1"
        assert format_num(10) == "10"
        assert format_num(0.000123456) == "0.000123456"

    def test_format_rate(self):
        assert format_rate(0, 0) == "0"
        assert format_rate(0, None) == "0"
        assert format_rate(None, 0) == "0"
        assert format_rate(12345, 0) == "0"
        assert format_rate(12345, 10.0) == "1234"
        assert format_rate(12345, 100.0) == "123.5"
        assert format_rate(12345, 1000.0) == "12.35"
        assert format_rate(12345, 10000.0) == "1.234"
        assert format_rate(12345, 100000.0) == "0.123"
        assert format_rate(12345, 1000000.0) == "0.012"

    def test_coerce_str(self):
        assert coerce_str("1.2") == 1.2
        assert coerce_str("1") == 1
        assert type(coerce_str("1")) is int
        assert coerce_str("") == ""
        assert coerce_str("1.2.3") == "1.2.3"
        assert coerce_str(None) is None

    def test_parse_option_args(self):
        res = parse_option_args([])
        assert res == {}
        res = parse_option_args(["foo:bar", "baz:42"])
        assert res == {"foo": "bar", "baz": 42}

    def test_parse_args_from_str(self):
        arg_def = (
            ("min", float),
            ("max", float, None),
        )
        res = parse_args_from_str("1.23", arg_def)
        assert res == {"min": 1.23, "max": None}

        res = parse_args_from_str("1.23, 42", arg_def)
        assert res == {"min": 1.23, "max": 42.0}

        with pytest.raises(TypeError):
            parse_args_from_str(None, arg_def)

        with pytest.raises(ValueError, match=".*Extra arg.*"):
            parse_args_from_str("1.23, 42, 32", arg_def)

        with pytest.raises(ValueError, match=".*convert string to float: 'foo'.*"):
            parse_args_from_str("1.23, foo", arg_def)

        arg_def = (
            ("min", float, 1.0),
            ("max", float, None),
        )
        res = parse_args_from_str("1.23", arg_def)
        assert res == {"min": 1.23, "max": None}

        res = parse_args_from_str("", arg_def)
        assert res == {"min": 1.0, "max": None}, "'$name()' uses default args"

        arg_def = (
            ("name", str),
            ("hint", str, "test"),
        )
        res = parse_args_from_str("foo", arg_def)
        assert res == {"name": "foo", "hint": "test"}

        res = parse_args_from_str("", arg_def)
        assert res == {"name": "", "hint": "test"}

        res = parse_args_from_str(",", arg_def)
        assert res == {"name": "", "hint": ""}

        # Allow single and double quotes
        res = parse_args_from_str("'foo', ' bar '", arg_def)
        assert res == {"name": "foo", "hint": " bar "}

        res = parse_args_from_str('"foo", " bar "', arg_def)
        assert res == {"name": "foo", "hint": " bar "}

        arg_def = (
            ("name", str),
            ("amount", float),
            ("hint", str, "test"),
        )
        with pytest.raises(ValueError, match="Missing mandatory arg `amount.*"):
            parse_args_from_str("1.23", arg_def)

        # Allow $() macros
        arg_def = (("amount", float),)
        res = parse_args_from_str("$(def_amount)", arg_def)
        assert res == {"amount": "$(def_amount)"}

        # # Parse `$(context_key)` macros
        # ctx = {"base_url": "http://example.com", "def_amount": 7}
        # res = parse_args_from_str("$(base_url)/foo, $(def_amount)", arg_def, ctx)
        # assert res == {"name": "http://example.com/foo", "amount": 8, "hint": "test"}

    def test_log(self):
        from snazzy import colors_enabled, enable_colors, green, red

        assert not colors_enabled()
        assert red("error") == "error"
        assert green("ok") == "ok"

        enable_colors(True, True)
        assert red("error") == "\x1b[91merror\x1b[39m"
        assert green("ok") == "\x1b[32mok\x1b[39m"



# -*- coding: utf-8 -*-
# (c) 2020-2023 Martin Wendt and contributors; see https://github.com/mar10/stressor
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
import re
from abc import ABC, abstractmethod

from stressor.util import StressorError, assert_always, check_arg, parse_args_from_str

#: (set) all activities accept these arguments
common_args = set(
    (
        "activity",
        "assert_match",
        "assert_max_time",
        "debug",
        "mock_result",
        "monitor",
        "name",
    )
)


class ActivityError(StressorError):
    """Base for all errors that are explicitly raised by activities."""


class ActivityTimeoutError(ActivityError):
    """Activity timed out."""


class ActivityCompileError(ActivityError):
    """Raised when activity constructor fails."""


class ActivityAssertionError(ActivityError):
    """Assertion failed (e.g. `assert_match` argument, ...)."""

    def __init__(self, assertion_list):
        super().__init__("Activity assertion failed")
        check_arg(assertion_list, (str, list))
        if isinstance(assertion_list, str):
            assertion_list = [assertion_list]
        self.assertion_list = assertion_list


class ScriptActivityError(ActivityError):
    """Raised when a ScriptActivity fails."""


class ActivityBase(ABC):
    """
    Common base class for all activities.

    New activity plugins can be created by deriving from this class,
    importing it, and then calling :func:`register_activity_plugins()`.

    Names of derived classes should end with '...Activity'.
    Classes which names that begin with an underscore ('_') are ignored.
    """

    #: (str) Name by which the actvity can be used in YAML configurations.
    #: Defaults to class name without trailing `...Activity`, e.g.
    #: 'SleepActivity' is called as ``activity: Sleep``
    _script_name = None

    #: (set) If defined, the default implementation of `compile()` will raise
    #: an error if any of those args is not passed.
    _mandatory_args = None

    #: (set) If defined, the default implementation of `compile()` will raise
    #: an error if other args are passed to the constructor.
    _known_args = None

    #: (set) Internal use only!
    _all_known_args = None

    #: (tuple|None) List of argument names that will be displayed in path
    #: strings
    _info_args = None

    #: (bool)
    _default_monitor = False

    #: (bool)
    _default_ignore_timing = False

    @abstractmethod
    def __init__(self, config_manager, **activity_args):
        """
        The constructor is called by the :class:`ConfigManager`, when the
        configuration was read.

        The compiler replaces the activity name in the definition's `activity`
        attribute with this instance.
        The `execute()` method is then called by the session runner on or more
        times.
        The default implementation stores `activity_args` as `self.raw_args`.
        however derived classes may choose to add named args explicitly for
        better readability and checking.

        An implementation should

            - Check args and raise errors, so we get early load-time failures
              when the configuration is read and compiled.
            - Optionally cache some information in this instance that can be
              re-used in the execute calls.
            - Optionally call `compiler.add_warning()`

        Attributes:
            compile_path_short (str):
                Path to location of definition in the configuration, e.g.
                '/main/2/PutRequest'
            compile_path (str):
                Path to location of definition in the configuration with
                added argument infos, e.g.
                '/main/2/PutRequest($(base_url)/test.html)'
            raw_args (dict):
            monitor (bool):
            ignore_timing (bool):
        Args:
            compile_path (:class:`PathStack`):
                The location breadcrumb path of this activity in the configration
                structure, e.g. 'main/#4'.
            activity_args (dict):
                named arguments that were passed to the activity definition.
                Note that the arguments are read at load-time and are not yet
                expanded (i.e. may contain `$(context_var)` macros).
        """
        self.raw_args = activity_args
        self.compile_path_short = config_manager.stack.get_path(
            skip_segs=2, last_seg=self.get_script_name()
        )
        self.compile_path = config_manager.stack.get_path(
            skip_segs=2, last_seg=self.get_info()
        )
        passed = set(activity_args.keys())
        if self._mandatory_args:
            missing = self._mandatory_args.difference(passed)
            if missing:
                raise ActivityCompileError(
                    "Missing mandatory arguments: {}".format(missing)
                )

        if self._all_known_args is None:
            if self._known_args:
                self._all_known_args = common_args | self._known_args
            else:
                self._all_known_args = common_args

        extra = passed.difference(self._all_known_args)
        if extra:
            raise ActivityCompileError("Unsupported arguments: {}".format(extra))

        self.monitor = activity_args.get("monitor", self._default_monitor)
        self.ignore_timing = activity_args.get(
            "ignore_timing", self._default_ignore_timing
        )

    def __str__(self):
        """Return a descriptive string."""
        return self.get_info()

    #: Provide nicer display for pprint(), etc.
    __repr__ = __str__

    @classmethod
    def get_script_name(cls):
        # Note: we must check `cls.__dict__` instead of `cls._script_name`,
        # because we want to test the local class attribute (not a derived one):
        if cls.__dict__.get("_script_name") is None:
            assert_always(cls.__name__.endswith("Activity"))
            cls._script_name = cls.__name__[:-8]
        return cls._script_name

    def get_info(self, info_args=True, expanded_args=None, session=None):
        """Return a descriptive string (optionally using expanded args).

        This method can be called with or without the additional context of
        the current session and expanded argument macros.

        Args:
            info_args (tuple|bool):
                List of argument names that should be added to display string.
                True: default to `self._info_args` (see `None` by default)
                False: don't add arguments
                None: all arguments
            expanded_args (dict, optional):
                optional argment dict (defaults to `self.raw_args`)
            session (SessionManager, optional):
                Running session context
        """
        if info_args is True:
            info_args = self._info_args
        elif info_args is False:
            info_args = []
        arg_dict = self.raw_args if expanded_args is None else expanded_args
        if info_args:
            # Add selected args
            args = (
                "{}={!r}".format(a, arg_dict.get(a)) for a in info_args if a in arg_dict
            )
        else:  # add all args
            args = (
                "{}={!r}".format(*kv) for kv in arg_dict.items() if kv[0] != "activity"
            )
        return "{}({})".format(self.get_script_name(), ", ".join(args))

    def prepare_execute(self, session, expanded_args):
        """Allow an activity to prepare the next execution.

        The session manager calls this for every activity instance, directly
        before `get_info()` and `execute()`.
        Normally this method does not need to be implemented. (One exception is
        `SleepActivity`, that calculates the next random duration per session,
        so it can be displayed by `get_info(..., session)`.)
        """
        pass

    @abstractmethod
    def execute(self, session, **expanded_args):
        """
        Called by the :class:`SessionManager`, after `$(context.var)` macros
        have been resolved if any.
        A derived class MUST implement this method.

        The session manager calls this methods for every activity instance:

        1. prepare_execute()
        2. get_info()
        3. execute()

          - Call session.log_warning()
          - raise ActivityError() for errors that a user can ignore by --force flag
          - Honor session.stop_request
          - Honor session.dry_run

        Args:
            session (:class:`SessionManager`):
            expanded_args (dict):
                current global vars
        Raises:
            ActivityError:
            Exception: all other exceptions are considered a fatal error
        Returns:
            nothing
        """


class MacroBase(ABC):
    """
    Common base class for all load-time script macros of the form ``$NAME(ARGS)``.

    These macros are applied at load time to the nested dict that was read from
    a YAML file.
    Typically they replace the value of that dict element (``parent[parent_key]``)
    with new content.

    For example

      - ``$load(PATH)`` replaces the element with a parsed YAML content or
        the text content of a python file.
      - ``$sleep(DURATION)`` replaces the element with a :class:`SleepActivity`
        definition

    Note that context variable expansions of the form ``$(CONTEXT.VAR.NAME)``
    are *not* handled by this kind of macro, because this has to be dealt with
    at run time.

    TODO: This means that ARGS cannot be dynamic, so for example $load($(my_file_name))
    will not work.

    Custom macro plugins can be defined by deriving from this class, importing
    it, and then calling :func:`register_macro_plugins()`.
    """

    #: Name for use in scripts.
    #: Defaults to lowercase class name without trailing `...Macro`, e.g.
    #: 'LoadMacro' is exposed as ``$load(...)``
    _script_name = None
    #: Regular expression that extracts '$NAME(ARGS)'
    #: Defaults to ...
    _regex = None
    #: List of tuples that define the supported positional arguments and
    #: optional default values.
    #: See :func:`stressor.util.parse_args_from_str` for syntax.
    #: If `None`, the raw cotent inside the brackets is passed as single string .
    _args_def = None

    #: Allow late evaluation (e.g. $stamp)
    #: TODO: Not yet implemented
    run_time_eval = False

    def __init__(self, **macro_args):
        """"""

    @classmethod
    def get_script_name(cls):
        # Note: we must check `cls.__dict__` instead of `cls._script_name`,
        # because we want to test the local class attribute (not a derived one):
        if cls.__dict__.get("_script_name") is None:
            assert_always(cls.__name__.endswith("Macro"))
            cls._script_name = cls.__name__[:-5].lower()
        return cls._script_name

    @classmethod
    def get_regex(cls):
        if cls._regex is None:
            name = cls.get_script_name()
            cls._regex = re.compile(r"\s*\$" + name + r"\((.*)\)\s*")
        return cls._regex

    def match_apply(self, context_reader, parent, parent_key):
        """Parse current value and call self.apply() if the macro matches.

        This default implementation supports
        Returns:
            (bool, any) (handled, )
        """
        pattern = self.get_regex()
        value = parent[parent_key]
        match = pattern.match(value)
        if not match:
            return (False, None)
        args = match.groups()
        assert len(args) == 1
        arg_str = args[0]
        if self._args_def:
            kwargs = parse_args_from_str(arg_str, self._args_def)
            res = self.apply(context_reader, parent, parent_key, **kwargs)
        else:
            res = self.apply(context_reader, parent, parent_key, arg_str)
        # logger.debug("Eval {}: {} => {}".format(stack, value, res))
        return (True, res)

    @abstractmethod
    def apply(self, context_reader, parent, parent_key, args_str):
        """
        Args:
            context_reader (:class:`ConfigManager`):
            parent (dict|list):
            parent_key (str|int):
            args_str (str|**kwargs):
        Returns:
            (any) The result that was produced (and stored into `parent[parent_key]`)
        """



# -*- coding: utf-8 -*-
# (c) 2020-2023 Martin Wendt and contributors; see https://github.com/mar10/stressor
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
https://w3c.github.io/web-performance/specs/HAR/Overview.html
"""
import json
import logging
import os
import re
from collections import Counter
from operator import itemgetter
from pprint import pformat
from urllib.parse import urlparse

from stressor.util import (
    base_url,
    datetime_to_iso,
    format_elap,
    is_yaml_keyword,
    iso_to_stamp,
    lstrip_string,
    shorten_string,
)

logger = logging.getLogger("stressor.har")

EMPTY_TUPLE = tuple()
PATTERN_TYPE = type(re.compile("foo"))  # 're.Pattern' requires Python 3.7+


class HarConverter:
    """Convert HAR file to YAML scenario."""

    #: Defaults for optional `--opts` paramter.
    DEFAULT_OPTS = {
        "fspec": None,
        "target_folder": None,
        "force": False,
        "base_url": True,  # True: automatic
        "collate_max_len": 100,  # put max 20 URL into one bucket
        "collate_max_duration": 1.0,  # a bucket spans max. 5 seconds
        "collate_max_pause": 0.2,  # 1 second gap will trigger open a new bucket
        "collate_thread_count": 5,
        # "add_req_comments": True,
        "skip_externals": True,
        "statics_types": (
            re.compile("image/.*", re.IGNORECASE),
            re.compile(".*/css", re.IGNORECASE),
            re.compile(".*/javascript", re.IGNORECASE),
            # "text/html",
        ),
        "skip_errors": True,
    }

    def __init__(self, opts):
        # self.stats = Counter()
        self.page_map = {}
        self.har_version = None
        self.browser_info = None
        self.creator_info = None
        self.first_entry_dt = None
        self.entries = []
        self.stats = Counter()
        self.prefix_counter = Counter()
        self.encoding = "utf-8-sig"
        self.opts = self.DEFAULT_OPTS.copy()
        self.opts.update(opts)
        #: (str) Available after `self._parse()`
        self.base_url = None
        self.polling_requests = []
        self.static_resources = []
        pl = []
        for pattern in self.opts["statics_types"]:
            if isinstance(pattern, str):
                pattern = re.compile(pattern, re.IGNORECASE)
            assert isinstance(pattern, PATTERN_TYPE)
            pl.append(pattern)
        self.opts["statics_types"] = pl

    def run(self):
        har_path = self.opts.get("fspec")
        target_folder = self.opts.get("target_folder")

        if har_path is not None:
            har_path = os.path.abspath(har_path)
            if not os.path.isfile(har_path):
                raise FileNotFoundError(har_path)

            logger.info("Parsing {}...".format(har_path))
            self._parse(har_path)

            self._postprocess()

        if target_folder is None:
            assert har_path
            name = os.path.splitext(os.path.basename(har_path))[0].strip(".")
            target_folder = "./{}".format(name)
        target_folder = os.path.abspath(target_folder)
        if not os.path.isdir(target_folder):
            logger.info("Creating folder {}...".format(target_folder))
            os.mkdir(target_folder)

        self._init_from_templates(target_folder)

        if har_path is not None:
            fspec = os.path.join(target_folder, "main_sequence.yaml")
            self._write_sequence(fspec)
        return True

    def _copy_template(self, tmpl_name, target_path, kwargs):
        if not self.opts["force"] and os.path.isfile(target_path):
            raise RuntimeError(
                "File exists (use --force to continue): {}".format(target_path)
            )
            # raise FileExistsError(target_path)
        tmpl_folder = os.path.dirname(__file__)
        src_path = os.path.join(tmpl_folder, tmpl_name)
        tmpl = open(src_path, "rt", encoding=self.encoding).read()
        tmpl = tmpl.format(**kwargs)
        logger.info("Writing {:,} bytes to {!r}...".format(len(tmpl), target_path))
        with open(target_path, "w") as fp:
            fp.write(tmpl)
        # print(tmpl)
        return

    def _init_from_templates(self, target_folder):
        ctx = {
            "date": datetime_to_iso(),
            "name": "NAME",
            "tag": "TAG",
            "base_url": self.base_url,
            "details": "",
        }
        self._copy_template(
            "users.yaml.tmpl", os.path.join(target_folder, "users.yaml"), ctx
        )
        self._copy_template(
            "sequence.yaml.tmpl", os.path.join(target_folder, "main_sequence.yaml"), ctx
        )
        self._copy_template(
            "scenario.yaml.tmpl", os.path.join(target_folder, "scenario.yaml"), ctx
        )

    def _is_static(self, entry):
        mime_type = entry["resp_type"]
        for pattern in self.opts["statics_types"]:
            if pattern.match(mime_type):
                return True
        return False

    _time_names = ("blocked", "dns", "connect", "send", "wait", "receive")

    def _calc_request_time(self, har_entry):
        t = 0.0
        timings = har_entry["timings"]
        for name in self._time_names:
            v = timings.get(name)
            if v is not None and v > 0:
                t += v
        return t * 0.001

    def _add_entry(self, har_entry):
        """Store the most important properties of an HAR entry."""
        req = har_entry["request"]
        resp = har_entry["response"]
        content = resp.get("content")

        # Record the usage count of the scheme+netloc, so we can pick the best
        # base_url later:
        url = req["url"]
        parse_result = urlparse(url)
        prefix = "{o.scheme}://{o.netloc}".format(o=parse_result)
        self.prefix_counter[prefix] += 1

        entry_dt = har_entry["startedDateTime"]
        if not self.first_entry_dt:
            self.first_entry_dt = entry_dt

        entry = {
            "start": iso_to_stamp(entry_dt),
            "method": req["method"],
            "url": url,
            "elap": self._calc_request_time(har_entry),
        }
        if req.get("httpVersion").upper() not in ("", "HTTP/1.1"):
            logger.warning("Unknown httpVersion: {!r}".format(req.get("httpVersion")))
            # logger.warning("Unknown httpVersion: {}".format(req))

        if req.get("comment"):
            entry["comment"] = req["comment"]

        if req.get("queryString"):  # List of (name, value) tuples
            entry["query"] = req["queryString"]

        if req.get("postData"):
            post_data = req["postData"]
            if post_data.get("mimeType"):
                entry["mimeType"] = post_data.get("mimeType")

            if post_data.get("params"):
                entry["data"] = post_data.get("params")
            elif post_data.get("text"):
                # according to spec, 'text' should be mutually exclusive to 'params'
                # but Chrome produces both or sometimes text only:
                entry["data"] = post_data.get("text")

        if resp.get("comment"):
            entry["resp_comment"] = resp["comment"]
        if content.get("mimeType"):
            entry["resp_type"] = content["mimeType"]
        if content.get("size"):
            entry["resp_size"] = content["size"]

        values = req.get("cookies")
        if values:
            cookies = entry["cookies"] = []
            for val in values:
                cookies.append((val["name"], val["value"]))

        values = req.get("headers")
        if values:
            headers = entry["headers"] = []
            for val in values:
                headers.append((val["name"], val["value"]))

        self.entries.append(entry)

    def _parse(self, fspec):
        with open(fspec, "r", encoding=self.encoding) as fp:
            har_data = json.load(fp)
        log = har_data["log"]
        assert len(har_data.keys()) == 1

        self.har_version = log["version"]

        creator = log["creator"]
        self.creator_info = "{}/{}".format(creator["name"], creator["version"])

        browser = log.get("browser")
        if browser:
            self.browser_info = "{}/{}".format(browser["name"], browser["version"])

        for page in log.get("pages", EMPTY_TUPLE):
            self.page_map[page["id"]] = page

        for entry in log["entries"]:
            self._add_entry(entry)

        # Sort entries
        # ("However the reader application should always make sure the array is sorted")
        self.entries.sort(key=itemgetter("start"))

        logger.debug("HAR:\n{}".format(pformat(self.entries)))
        # print("HAR:\n{}".format(pformat(self.entries)))
        return

    def _postprocess(self):
        # Figure out base_url
        base_url = self.opts.get("base_url")
        if base_url is True:
            # The most-used scheme+netloc is used as base_url
            # print(pformat(self.prefix_counter))
            base_url = self.prefix_counter.most_common(1)[0][0]
        self.base_url = base_url
        logger.info("Using base_url {!r}.".format(base_url))
        # print(base_url)

        # Remove unwanted entries and strip base_url where possible
        el = []
        skip_ext = self.opts["skip_externals"]
        collate_max_len = self.opts["collate_max_len"]
        collate_max_duration = self.opts["collate_max_duration"]
        collate_max_pause = self.opts["collate_max_pause"]
        bucket = []
        bucket_start_stamp = 0
        bucket_last_stamp = 0

        def _flush():
            nonlocal bucket, bucket_start_stamp, bucket_last_stamp
            if not bucket:
                return
            self.stats["collated_activities"] += 1
            entry = bucket[0]
            entry["url_list"] = [e["url"] for e in bucket]
            # print("FLUSH\n{}".format(pformat(entry)))
            bucket = []
            bucket_start_stamp = 0
            bucket_last_stamp = 0

        for entry in self.entries:
            self.stats["entries_total"] += 1
            skip = False

            # Fix URL by removing the `base_url` prefix
            url = entry["url"]
            rel_url = lstrip_string(url, base_url, ignore_case=True)
            # The URL did not start with base_url, so it is 'external'
            is_external = rel_url is url
            if is_external:
                if skip_ext:
                    logger.warning("Skipping external URL: {}".format(url))
                    self.stats["external_urls"] += 1
                    skip = True
            else:
                entry["url_org"] = url
                entry["url"] = rel_url

            # Collate bursts of simple GET request to one single entry
            start = entry["start"]
            if len(bucket) >= collate_max_len:
                _flush()
            first = len(bucket) == 0

            if (
                collate_max_len > 1  # Collation enabled
                and entry["method"] == "GET"  # Only GET requests
                and not entry.get("data")  # No POST data
                and len(bucket) < collate_max_len  # Bucket size limit
                and (first or (start - bucket_start_stamp) <= collate_max_duration)
                and (first or (start - bucket_last_stamp) <= collate_max_pause)
                and self._is_static(entry)  # Only static resources (JS, CSS, ...)
            ):
                self.stats["collated_urls"] += 1
                bucket.append(entry)
                if not bucket_start_stamp:
                    bucket_start_stamp = start
                bucket_last_stamp = start
                # We keep the first entry, so we can add the bucket URLs there later
                if len(bucket) > 1:
                    skip = True
            elif bucket:
                # If a burst of simple requests is interrupted, flush and restart
                _flush()

            if skip:
                self.stats["skipped"] += 1
            else:
                self.stats["entries"] += 1
                el.append(entry)

        _flush()
        self.entries = el
        return

    activity_map = {
        "GET": "GetRequest",
        "PUT": "PutRequest",
        "POST": "PostRequest",
        "DELETE": "DeleteRequest",
    }

    def _write_args(self, lines, name, data):
        """
        Args:
            lines (list):
            name (str):
            data (list): a 'params' list of {name: ..., value: ...} objects
        """
        assert isinstance(data, (list, tuple))
        used = set()
        lines.append("  {}:\n".format(name))
        for p in data:
            # TODO: coerce value (which in HAR is always a string)

            # Convert [{name: ..., value: ...}, ...] syntax to dict
            # TODO: We do this, because activity code doesn not handle the
            #       tuple syntax yet. But we should fall back to tuples syntax
            #       if the list does contain duplicate names, instead of
            #       dicarding.
            name, value = p["name"], p["value"]
            # lines.append("    - ['{}', {}]\n".format(name, json.dumps(value)))

            if name in used:
                logger.error(
                    "Discarding multiple param name: {}: {}".format(name, value)
                )
                continue
            used.add(name)
            if not is_yaml_keyword(name):
                name = '"' + name + '"'
            lines.append("    {}: {}\n".format(name, json.dumps(value)))

        return

    def _write_entry(self, fp, entry):
        opts = self.opts
        activity = self.activity_map.get(entry["method"], "HTTPRequest")
        url_list = entry.get("url_list")
        is_bucket = bool(url_list) and len(url_list) > 1
        if is_bucket:
            activity = "StaticRequests"

        lines = []
        if entry.get("comment"):
            lines.append("# {}\n".format(shorten_string(entry["comment"], 75)))
        if is_bucket:
            lines.append("# Auto-collated {:,} GET requests\n".format(len(url_list)))
        else:
            lines.append(
                "# Response type: {!r}, size: {:,} bytes, time: {}\n".format(
                    entry.get("resp_type"),
                    entry.get("resp_size", -1),
                    format_elap(entry.get("elap"), high_prec=True),
                )
            )
        if entry.get("resp_comment"):
            lines.append("# {}\n".format(shorten_string(entry["resp_comment"], 75)))

        url = entry["url"]
        expand_url_params = False

        lines.append("- activity: {}\n".format(activity))
        if is_bucket:
            lines.append("  thread_count: {}\n".format(opts["collate_thread_count"]))
            lines.append("  url_list:\n")
            for url in url_list:
                lines.append("    - '{}'\n".format(url))
        else:
            if entry.get("query"):
                expand_url_params = True
                url = base_url(url)
            lines.append("  url: '{}'\n".format(url))

        if activity == "HTTPRequest":
            lines.append("  method: {}\n".format(entry["method"]))

        # TODO:
        # We have ?query also as part of the URL
        # if entry.get("query"):
        #     lines.append("  params: {}\n".format(entry["query"]))
        if expand_url_params:
            self._write_args(lines, "params", entry["query"])

        # TODO: set headers = {'Content-type': 'content_type_value'}
        #       if entry.dData.mimeType is defined

        data = entry.get("data")
        if data:
            if isinstance(data, (list, tuple)):
                # Must be a 'params' list of {name: ..., value: ...} objects
                self._write_args(lines, "data", data)
            else:
                assert type(data) is str
                logger.warning("Expected list, but got text: {!r}".format(data))
                lines.append("  data: {}\n".format(json.dumps(data)))

        lines.append("\n")
        fp.writelines(lines)

    def _write_sequence(self, fspec):
        logger.info("Writing activity sequence to {!r}...".format(fspec))
        with open(fspec, "wt") as fp:
            fp.write("# Stressor Activity Definitions\n")
            fp.write("# See https://stressor.readthedocs.io/\n")
            fp.write("# Auto-generated {}\n".format(datetime_to_iso()))
            fp.write("# Source:\n")
            fp.write("#     File: {}\n".format(self.opts["fspec"]))
            fp.write("#     HAR Version: {}\n".format(self.har_version))
            fp.write("#     Creator: {}\n".format(self.creator_info))
            if self.browser_info:
                fp.write("#     Browser: {}\n".format(self.browser_info))
            fp.write("#     Recorded: {}\n".format(self.first_entry_dt))
            fp.write("#     Using base URL {!r}\n".format(self.base_url))
            fp.write("\n")

            for entry in self.entries:
                self._write_entry(fp, entry)
        logger.info("Done.")
        return



# -*- coding: utf-8 -*-
# (c) 2020-2023 Martin Wendt and contributors; see https://github.com/mar10/stressor
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
import os

from stressor.run_manager import RunManager
from stressor.session_manager import User


class TestRunManager:
    def setup_method(self):
        self.fixtures_path = os.path.join(os.path.dirname(__file__), "fixtures")

    def test_user(self):
        user = User("Joe", "secret")
        assert user.auth == ("Joe", "secret")

    def test_dry_run(self):
        config_path = os.path.join(self.fixtures_path, "test_dry_run.yaml")
        rm = RunManager()

        activities = []
        errors = []

        def notify_hook(channel, *args, **kwargs):
            if channel == "end_activity":
                activities.append(
                    "{} {}".format(kwargs.get("path"), kwargs.get("activity_args"))
                )
                print(
                    "EXECUTE {} {}".format(
                        kwargs.get("path"), kwargs.get("activity_args")
                    )
                )

            if kwargs.get("error"):
                errors.append("{} {error}".format(channel, **kwargs))

            # print("notify_hook({})".format(channel), args, kwargs)
            # print("notify_hook({})".format(channel), kwargs.get("path"), kwargs.get("elap"))

        rm.subscribe("*", notify_hook)
        rm.load_config(config_path)
        # assert rm.run_config
        assert rm.config_manager.file_version == 0
        # assert rm.c"extra_opt".file_version == 0
        # assert "GetRequest" in activity_plugin_map

        options = {}
        extra_config = {
            "extra_opt": "extra_val",
            "dry_run": True,
        }
        res = rm.run(options, extra_config)
        assert res is True
        assert errors == []

        si = rm.get_status_info()
        assert si["name"] == "test_dry_run"
        assert si["hasErrors"] == 0

        # assert activities == []
        # assert 0

    def test_mock_server(self, mock_wsgidav_server_fixture):
        config_path = os.path.join(self.fixtures_path, "test_mock_server.yaml")
        rm = RunManager()
        rm.load_config(config_path)
        # assert rm.run_config
        assert rm.config_manager.file_version == 0
        options = {}
        extra_config = {}
        res = rm.run(options, extra_config)
        assert res is True
        # assert 0



# -*- coding: utf-8 -*-
# (c) 2020-2023 Martin Wendt and contributors; see https://github.com/mar10/stressor
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""

import argparse

# --- verbose_parser ----------------------------------------------------------

verbose_parser = argparse.ArgumentParser(
    add_help=False,
    # allow_abbrev=False,
)

qv_group = verbose_parser.add_mutually_exclusive_group()
qv_group.add_argument(
    "-v",
    "--verbose",
    action="count",
    default=3,
    help="increment verbosity by one (default: %(default)s, range: 0..5)",
)
qv_group.add_argument(
    "-q", "--quiet", default=0, action="count", help="decrement verbosity by one"
)


# --- common_parser ----------------------------------------------------------


common_parser = argparse.ArgumentParser(
    add_help=False,
    # allow_abbrev=False,
)
common_parser.add_argument(
    "-n",
    "--dry-run",
    action="store_true",
    help="just simulate and log results, but don't change anything",
)
common_parser.add_argument(
    "--no-color", action="store_true", help="prevent use of ansi terminal color codes"
)
common_parser.add_argument(
    "--log",
    dest="log_file",
    help="Path to log file or folder (generate unique file name in the latter case)",
)



# -*- coding: utf-8 -*-
# (c) 2020-2023 Martin Wendt and contributors; see https://github.com/mar10/stressor
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
import logging
import os
import platform
import random
import re
import sys
import types
import warnings
from datetime import datetime
from urllib.parse import urljoin, urlparse

from dateutil.parser import isoparse

from stressor import __version__

logger = logging.getLogger("stressor")

# Use logger.important("message even in -q mode")
LOGLEVEL_IMPORTANT = logging.WARNING + 1
logging.addLevelName(LOGLEVEL_IMPORTANT, "NOTE")


def log_important(self, message, *args, **kws):
    if self.isEnabledFor(LOGLEVEL_IMPORTANT):
        # Yes, logger takes its '*args' as 'args'.
        self._log(LOGLEVEL_IMPORTANT, message, args, **kws)


logging.Logger.important = log_important


#: Check if a a string may be used as YAML dictionary key without using quotes.
#: NOTE: YAML evaluates `0_` as "0" and `0_1_` as "1", so we don't accept leading numbers
RE_YAML_KEYWORD = re.compile(r"^[a-zA-Z_]+\w*$")

PYTHON_VERSION = "{}.{}.{}".format(
    sys.version_info[0], sys.version_info[1], sys.version_info[2]
)
version_info = "stressor/{} Python/{}({} bit) {}".format(
    __version__,
    PYTHON_VERSION,
    "64" if sys.maxsize > 2**32 else "32",
    platform.platform(),
)


class StressorError(RuntimeError):
    """Base class for all exception that we deliberatly throw."""


class NO_DEFAULT:
    """Used as default parameter to distinguish from `None`."""


# class Singleton(type):
#     """
#     See also https://stackoverflow.com/a/6798042/19166

#     Example::

#         class Logger(metaclass=Singleton):
#             pass

#     """

#     _instances = {}

#     def __call__(cls, *args, **kwargs):
#         if cls not in cls._instances:
#             cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
#         return cls._instances[cls]


class _VOID_SEGMENT:
    """Used internally."""


class PathStack:
    """
    Examples::

        path = PathStack()
        with path.enter("level_1"):
            print("Current location: {}".format(path))
    """

    def __init__(self, root_name=None, delimiter="/"):
        self.stack = []
        self.delimiter = delimiter
        if root_name:
            self.push(root_name)

    def __str__(self):
        return self.get_path()
        # stack = (str(s) for s in self.stack if s is not _VOID_SEGMENT)
        # return self.delimiter + self.delimiter.join(stack)

    #: Provide nicer display for pprint(), etc.
    def __repr__(self):
        return f"PathStack<{self}>"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.pop()

    def enter(self, name, ignore=False):
        if ignore:
            self.push(_VOID_SEGMENT)
        else:
            self.push(name)
        return self

    def level(self):
        return len(self.stack)

    def push(self, name):
        self.stack.append(name)

    def pop(self):
        return self.stack.pop()

    def get_path(self, skip_segs=0, last_seg=None):
        stack = [str(s) for s in self.stack if s is not _VOID_SEGMENT]
        if skip_segs > 0:
            stack = stack[skip_segs:]
        if last_seg is not None:
            stack = stack[:-1]
            stack.append(last_seg)
        return self.delimiter + self.delimiter.join(stack)


def timetag(seconds=True, ms=False):
    """Return a time stamp string that can be used as (part of a) filename (also sorts well)."""
    now = datetime.now()
    if ms or seconds:
        s = now.strftime("%Y%m%d_%H%M%S")
        if ms:
            s = "{}_{}".format(s, now.microsecond)
    else:
        s = now.strftime("%Y%m%d_%H%M")
    return s


def check_cli_verbose(default=3):
    """Check for presence of `--verbose`/`--quiet` or `-v`/`-q` without using argparse."""
    args = sys.argv[1:]
    verbose = default + args.count("--verbose") - args.count("--quiet")

    for arg in args:
        if arg.startswith("-") and not arg.startswith("--"):
            verbose += arg[1:].count("v")
            verbose -= arg[1:].count("q")
    return verbose


def init_logging(verbose=3, path=None):
    """CLI calls this."""
    if verbose < 1:
        level = logging.CRITICAL
    elif verbose < 2:
        level = logging.ERROR
    elif verbose < 3:
        level = logging.WARNING
    elif verbose < 4:
        level = logging.INFO
    else:
        level = logging.DEBUG

    logging.basicConfig(
        level=level,
        # format="%(asctime)s.%(msecs)03d <%(thread)d> %(levelname)-7s %(message)s",
        format="%(asctime)-8s.%(msecs)-3d <%(thread)05d> %(levelname)-7s %(message)s",
        # format="%(asctime)s.%(msecs)03d <%(process)d.%(thread)d> %(levelname)-8s %(message)s",
        # fmt='%(asctime)s.%(msecs)03d',datefmt='%Y-%m-%d,%H:%M:%S')
        # format="%(asctime)s.%(msecs)d <%(process)d.%(thread)d> %(message)s",
        datefmt="%H:%M:%S",
    )
    # Make sure it is set (when running from tox, the above basicConfig() did not set it)
    logger.setLevel(level)

    if path:
        if os.path.isdir(path):
            fname = "stressor_{}.log".format(timetag())
            path = os.path.join(path, fname)
        logger.info("Writing log to '{}'".format(path))
        if os.path.isfile(path):
            logger.warning("Removing log file '{}'".format(path))
            os.remove(path)
        hdlr = logging.FileHandler(path)
        formatter = logging.Formatter(
            "%(asctime)s.%(msecs)-3d - %(levelname)s: %(message)s", "%H:%M:%S"
        )
        hdlr.setFormatter(formatter)
        logger.addHandler(hdlr)
        # logger.setLevel(logging.DEBUG)
        logger.info("Start log ({})".format(datetime.now()))
        logger.info(version_info)
        logger.info("Running {}".format(" ".join(sys.argv)))

        # redirect `logger` to our special log file as well:
        logger.addHandler(hdlr)

    # Silence requests `InsecureRequestWarning` messages
    if verbose < 3:
        warnings.filterwarnings("ignore", message="Unverified HTTPS request")
    else:
        warnings.filterwarnings("once", message="Unverified HTTPS request")

    return logger


def set_console_ctrl_handler(
    ctrl_handler, do_ctrl_c=True, do_ctrl_break=False, do_ctrl_close=False
):
    """

    The do_ctrl_* functions could simply be sys.exit(1), which will ensure that
    atexit handlers get called.
    See https://bugs.python.org/issue35935

    Raises:
        ctypes.WinError
    Returns:
        False if not on Windows or could not register the handler.
    """
    if os.name != "nt":
        return False

    try:
        import ctypes
        from ctypes import wintypes

        kernel32 = ctypes.WinDLL("kernel32", use_last_error=True)
        logger.info("Loaded kernel32 DLL on Windows.")
    except Exception as e:
        logger.warning("Could not load kernel32 DLL on Windows: {}".format(e))
        return False

    CTRL_C_EVENT = 0
    CTRL_BREAK_EVENT = 1
    CTRL_CLOSE_EVENT = 2

    HANDLER_ROUTINE = ctypes.WINFUNCTYPE(wintypes.BOOL, wintypes.DWORD)
    kernel32.SetConsoleCtrlHandler.argtypes = (HANDLER_ROUTINE, wintypes.BOOL)

    @HANDLER_ROUTINE
    def handler(ctrl):
        if ctrl == CTRL_C_EVENT and do_ctrl_c:
            handled = ctrl_handler(ctrl)
        elif ctrl == CTRL_BREAK_EVENT and do_ctrl_break:
            handled = ctrl_handler(ctrl)
        elif ctrl == CTRL_CLOSE_EVENT and do_ctrl_close:
            handled = ctrl_handler(ctrl)
        else:
            handled = False
        # If not handled, call the next handler.
        return handled

    if not kernel32.SetConsoleCtrlHandler(handler, True):
        raise ctypes.WinError(ctypes.get_last_error())
    return True


def assert_always(condition, msg=None):
    """`assert` even in production code."""
    try:
        if not condition:
            raise AssertionError(msg) if msg is not None else AssertionError
    except AssertionError as e:
        if sys.version_info < (3, 7):
            raise
        # Strip last frames, so the exception's stacktrace points to the call
        # Credits: https://stackoverflow.com/a/58821552/19166
        _exc_type, _exc_value, traceback = sys.exc_info()
        back_frame = traceback.tb_frame.f_back

        back_tb = types.TracebackType(
            tb_next=None,
            tb_frame=back_frame,
            tb_lasti=back_frame.f_lasti,
            tb_lineno=back_frame.f_lineno,
        )
        raise e.with_traceback(back_tb)


def _check_arg(argument, types, condition, accept_none):
    if __debug__:
        err_msg = "`allowed_types` must be a type or class (or a tuple thereof): got instance of {}"
        if isinstance(types, tuple):
            for t in types:
                assert isinstance(t, type), err_msg.format(type(t))
        else:
            assert isinstance(types, type), err_msg.format(type(types))

    if accept_none:
        if argument is None:
            return
        extra_msg = "`None` or "
    else:
        extra_msg = ""

    if not isinstance(argument, types):
        raise TypeError(
            "Expected {}{}, but got {}".format(extra_msg, types, type(argument))
        )
    if condition is not NO_DEFAULT and not bool(condition):
        raise ValueError(
            "Invalid argument value: {} {}".format(type(argument), argument)
        )


def check_arg(argument, allowed_types, condition=NO_DEFAULT, or_none=False):
    """Check if `argument` has the expected type and value.

    **Note:** the exception's traceback is manipulated, so that the back frame
    points to the ``check_arg()`` line, instead of the actual ``raise``.

    Args:
        argument (any): value of the argument to check
        allowed_types (type or tuple of types)
        condition (bool, optional): additional condition that must be true
        or_none (bool, optional): defaults to false
    Returns:
        None
    Raises:
        TypeError: if `argument` is of an unexpected type
        ValueError: if `argument` does not fulfill the optional condition
        AssertionError:
            if `check_arg` was called with a wrong syntax, i.e. `allowed_types`
            does not contain types, e.g. if `1` was passed instead of `int`.

    Examples::

        def foo(name, amount, options=None):
            check_arg(name, str)
            check_arg(amount, (int, float), amount > 0)
            check_arg(options, dict, or_none=True)
    """
    try:
        _check_arg(argument, allowed_types, condition, accept_none=or_none)
    except (TypeError, ValueError) as e:
        if sys.version_info < (3, 7):
            raise
        # Strip last frames, so the exception's stacktrace points to the call
        _exc_type, _exc_value, traceback = sys.exc_info()
        back_frame = traceback.tb_frame.f_back

        back_tb = types.TracebackType(
            tb_next=None,
            tb_frame=back_frame,
            tb_lasti=back_frame.f_lasti,
            tb_lineno=back_frame.f_lineno,
        )
        raise e.with_traceback(back_tb)


def get_dict_attr(d, key_path, default=NO_DEFAULT):
    """Return the value of a nested dict using dot-notation path.

    Args:
        d (dict):
        key_path (str):
    Raises:
        KeyError:
        ValueError:
        IndexError:

    Examples::

        ...

    Todo:
        * k[1] instead of k.[1]
        * default arg
    """
    if default is not NO_DEFAULT:
        try:
            return get_dict_attr(d, key_path)
        except (AttributeError, KeyError, ValueError, IndexError):
            return default

    check_arg(d, dict)

    seg_list = key_path.split(".")
    value = d[seg_list[0]]
    for seg in seg_list[1:]:
        if isinstance(value, dict):
            value = value[seg]
        elif isinstance(value, (list, tuple)):
            if not seg.startswith("[") or not seg.endswith("]"):
                raise ValueError("Use `[INT]` syntax to address list items")
            seg = seg[1:-1]
            value = value[int(seg)]
        else:
            # raise ValueError("Segment '{}' cannot be nested".format(seg))
            try:
                value = getattr(value, seg)
            except AttributeError:
                raise  # ValueError("Segment '{}' cannot be nested".format(seg))

    return value


def coerce_str(s):
    """Return `s` converted to float, int, or str."""
    check_arg(s, str, or_none=True)
    if s is None:
        return None
    for t in (int, float, str):  # order matters!
        try:
            return t(s)
        except ValueError:
            if t is str:
                raise


def parse_option_args(opt_list, coerce_values=True):
    """Evaluate `--option` args.

    Args:
        opt_list (list):
            List of option definitions of the form "OPT_NAME:OPT_VALUE"
        coerce_values (bool=true):
    Returns:
        Dict of {opt_name: opt_value, ...}
    """
    check_arg(opt_list, list, or_none=True)
    res = {}
    if not opt_list:
        return res
    for opt in opt_list:
        # print(opt_list, opt)
        if ":" not in opt:
            raise ValueError("Expected 'NAME:VALUE', got {!r}".format(opt))
        config_name, config_val = opt.split(":", 1)
        if coerce_values:
            res[config_name.strip()] = coerce_str(config_val)
        else:
            res[config_name.strip()] = config_val
    # print(opt_list, res)
    return res


def parse_args_from_str(arg_str, arg_defs):  # , context=None):
    """
    Args:
        args_str (str): argument string, optionally comma-separated
        arg_defs (tuple): list of argument definitions
        context (dict, optional):
            When passed, the arguments  are parsed for ``$(var_name)`` macros,
            to lookup values from that dict.
    Returns:
        (dict) keyword args
    Raises:
        TypeError: if `argument` is of an unexpected type
        ValueError: if `argument` does not fulfill the optional condition
        AssertionError:
            if `parse_args_from_str` was called with a wrong syntax, i.e.
            `arg_defs` is not well-formed.

    Examples::

        arg_defs = (
            ("name", str),
            ("amount", float),
            ("options", dict, {}),
        )
        def order(raw_arg_string):
            kwargs = parse_args_from_str(arg_defs)
            assert isisnstance(kwargs["name"], str)
            assert type(kwargs["amount"]) is float
            assert isisnstance(kwargs["options"], dict)
    """
    check_arg(arg_str, str)
    check_arg(arg_defs, (list, tuple))

    res = {}

    # Special case: '$name()' should not be interpreted as having one "" arg
    # if arg_defs defines a default for the first arg
    if arg_str.strip() == "" and len(arg_defs[0]) == 3:
        arg_str = str(arg_defs[0][2])

    arg_list = [a.strip() for a in arg_str.split(",")]
    optional_mode = False
    for arg_def in arg_defs:
        check_arg(arg_def, (list, tuple))
        if len(arg_def) == 2:
            arg_name, arg_type = arg_def
            arg_default = NO_DEFAULT
            if optional_mode:
                raise AssertionError(
                    "Mandatory arg definition must not follow optional args: `{}`".format(
                        arg_def
                    )
                )
        elif len(arg_def) == 3:
            arg_name, arg_type, arg_default = arg_def
            optional_mode = True
        else:
            raise AssertionError("Expected 2- or 3-tuple: {}".format(arg_def))

        if arg_type not in (float, int, str):
            raise AssertionError(
                "Unsupported argument definition type: {}".format(arg_def)
            )

        try:
            # Get next arg
            arg_val = arg_list.pop(0)
            # Allow quotes
            is_quoted = (arg_val.startswith('"') and arg_val.endswith('"')) or (
                arg_val.startswith("'") and arg_val.endswith("'")
            )
            if is_quoted:
                # Strip quotes and return as string (don't cast to other types)
                arg_val = arg_val[1:-1]
            elif "$(" in arg_val:
                # The arg seems to be a macro: don't try to cast.
                pass
            else:
                # Raises ValueError:
                arg_val = arg_type(arg_val)
        except IndexError:
            if arg_default is NO_DEFAULT:
                raise ValueError(
                    "Missing mandatory arg `{}` in '{}'.".format(arg_name, arg_str)
                )
            arg_val = arg_default

        res[arg_name] = arg_val

    if arg_list:
        raise ValueError("Extra args `{}`.".format(", ".join(arg_list)))

    return res


def is_yaml_keyword(s):
    """Return true if `s` is a JSON compatible key."""
    return bool(s and RE_YAML_KEYWORD.match(s))


# def is_abs_url(url):
#     """Return true if url is already absolute."""
#     return "://" in url or url.startswith("/")


def is_relative_url(url):
    """Return true if url is relative to the server."""
    return "://" not in url  # or url.startswith("/")


def resolve_url(root, url):
    """Convert relative URL to absolute, using `root` as default."""
    return urljoin(root, url)


def base_url(url):
    parsed_uri = urlparse(url)
    if parsed_uri.netloc:
        res = "{uri.scheme}://{uri.netloc}{uri.path}".format(uri=parsed_uri)
    else:
        res = "{uri.path}".format(uri=parsed_uri)
    return res


def shorten_string(long_string, max_chars, min_tail_chars=0, place_holder="[...]"):
    """Return string, shortened to max_chars characters.

    long_string = "This is a long string, that will be truncated."
    trunacated_string = truncate_string(long_string, max_chars=26, min_tail_chars=11, place_holder="[...]")
    print trunacated_string
    >> This is a [...] truncated.

    @param long_string: string to be truncated
    @param max_chars: max chars of returned string
    @param min_tail_chars: minimum of tailing chars
    @param place_holder: place holder for striped content
    @return: truncated string
    """

    assert max_chars > (min_tail_chars + len(place_holder))

    # Other types aren't supported.
    if not isinstance(long_string, str):
        return long_string

    # If string is shorter then max_chars, it can be returned, as is.
    elif len(long_string) <= max_chars:
        return long_string

    # Making sure we don't exceed max_chars in total.
    prefix_length = max_chars - min_tail_chars - len(place_holder)

    if min_tail_chars:
        long_string = (
            long_string[:prefix_length] + place_holder + long_string[-min_tail_chars:]
        )
    else:
        long_string = long_string[:prefix_length] + place_holder

    assert len(long_string) == max_chars

    return long_string


def get_random_number(num_or_tuple):
    check_arg(num_or_tuple, (int, float, list, tuple), or_none=True)

    if isinstance(num_or_tuple, (tuple, list)):
        v = random.uniform(*num_or_tuple)
    elif num_or_tuple:
        v = float(num_or_tuple)
    else:
        v = num_or_tuple
    return v


def datetime_to_iso(dt=None, microseconds=False):
    """Return current UTC datetime as ISO formatted string."""
    if dt is None:
        dt = datetime.now()
    if not microseconds:
        dt.replace(microsecond=0)
    return dt.isoformat(sep=" ")


def iso_to_datetime(iso):
    """Convert as ISO formatted datetime string to datetime."""
    # dt = datetime.strptime(iso, "%Y-%m-%dT%H:%M:%S.%fZ")
    dt = isoparse(iso)

    return dt


def iso_to_stamp(iso):
    """Convert as ISO formatted datetime string to datetime."""
    return iso_to_datetime(iso).timestamp()


def format_elap(seconds, count=None, unit="items", high_prec=False):
    """Return elapsed time as H:M:S.h string with reasonable precision."""
    days, seconds = divmod(seconds, 86400) if seconds else (0, 0)

    if seconds >= 3600:
        m, s = divmod(seconds, 60)
        h, m = divmod(m, 60)
        if high_prec:
            res = "{:d}:{:02d}:{:04.1f} hrs".format(int(h), int(m), s)
        else:
            res = "{:d}:{:02d}:{:02d} hrs".format(int(h), int(m), int(s))
    elif seconds >= 60:
        m, s = divmod(seconds, 60)
        if high_prec:
            res = "{:d}:{:05.2f} min".format(int(m), s)
        else:
            res = "{:d}:{:02d} min".format(int(m), int(s))
    else:
        if high_prec:
            if seconds > 0.01:
                res = "{:.3f} sec".format(seconds)
            else:
                res = "{:f} sec".format(seconds)
        elif seconds > 5:
            res = "{:.1f} sec".format(seconds)
        else:
            res = "{:.2f} sec".format(seconds)

    if days == 1:
        res = "{} day {}".format(int(days), res)
    elif days:
        res = "{} days {}".format(int(days), res)

    if count and (seconds > 0):
        res += ", {:,.1f} {}/sec".format(float(count) / seconds, unit)
    return res


def format_num(num):
    """Return num rounded to reasonable precision (promille)."""
    if num >= 1000.0:
        res = "{:,}".format(round(num))
    elif num >= 100.0:
        res = str(round(num, 1))
        # res = "{:,.1f}".format(num)
    elif num >= 10.0:
        res = str(round(num, 2))
        # res = "{:,.2f}".format(num)
    elif num >= 1.0:
        res = str(round(num, 3))
        # res = "{:,.3f}".format(num)
    else:
        res = str(num)
        # res = "{:,}".format(num)
    return res


def format_rate(count, time, unit=None, high_prec=False):
    """Return count / time with reasonable precision."""
    if not time or not count:
        return "0"

    rate = float(count) / float(time)
    if rate >= 1000:
        res = "{}".format(int(round(rate)))
    elif rate >= 100:
        res = "{}".format(round(rate, 1))
    elif rate >= 10:
        res = "{}".format(round(rate, 2))
    else:
        res = "{}".format(round(rate, 3))
    return res


# def format_relative_datetime(dt, as_html=False):
#     """Format a datetime object as relative expression (i.e. '3 minutes ago')."""
#     try:
#         diff = datetime.utcnow() - dt
#         res = None
#         s = diff.seconds

#         if diff.days > 7 or diff.days < 0:
#             res = dt.strftime("%d %b %y")
#         elif diff.days == 1:
#             res = "1 day ago"
#         elif diff.days > 1:
#             res = "{} days ago".format(diff.days)
#         elif s <= 1:
#             res = "just now"
#         elif s < 60:
#             res = "{} seconds ago".format(s)
#         elif s < 120:
#             res = "1 minute ago"
#         elif s < 3600:
#             res = "{} minutes ago".format(s/60)
#         elif s < 7200:
#             res = "1 hour ago"
#         else:
#             res = "{} hours ago".format(s/3600)

#         if as_html:
#             dts = dt.strftime("%Y-%m-%d %H:%M:%S UTC")
#             res = "<span title='{dt}'>{rel}</span>".format(rel=res, dt=dts)

#     except Exception:
#         log_exception("invalid dt: {}".format(dt))
#         res = "(invalid {})".format(dt)
#     return res


# def format_relative_stamp(stamp, as_html=False):
#     dt = stamp_to_datetime(stamp)
#     return format_relative_datetime(dt, as_html)


# def string_to_bool(arg_, default=None):
#     """Return boolean for various bool string representations.

#     Raises Error, if string is not compatible, and no default was passed.
#     """
#     if not default in (None, True, False):
#         raise TypeError("default must be None or boolean")
#     if arg_ is None:
#         if default in (True, False):
#             return default
#         else: #default == None
#             raise ValueError("Argument is %r and default %r." % (arg_, default))
#     else:
#         if arg_ in (True, False):
#             # `bool` is a subclass of `int`, so we get here for 0 and 1 as well!
#             return bool(arg_)
#         arg = arg_.lower().strip()
#         if arg in ("1", "true", "on", "yes"):
#             return True
#         elif arg in ("0", "false", "off", "no"):
#             return False
#         elif default is not None:
#             return default
#     raise ValueError("Argument string is not boolean: %r default: %r." % (arg_, default))


# def byteNumberString(number, thousandsSep=True, partition=False,
#                      base1024=True, appendBytes="iec", prec=0):
#     """Convert bytes into human-readable representation."""
#     magsuffix = ""
#     bytesuffix = ""
#     assert appendBytes in (False, True, "short", "iec")
#     if partition:
#         magnitude = 0
#         if base1024:
#             while number >= 1024:
#                 magnitude += 1
# #                 number = number >> 10
#                 number /= 1024.0
#         else:
#             while number >= 1000:
#                 magnitude += 1
#                 number /= 1000.0
#         # TODO:
#         # Windows 7 Explorer teilt durch 1024, und verwendet trotzdem KB statt KiB.
#         # Außerdem wird aufgerundet: --> 123 -> "1 KB"
#         # http://en.wikipedia.org/wiki/Kibi-#IEC_standard_prefixes
#         magsuffix = ["", "K", "M", "G", "T", "P"][magnitude]
#         if magsuffix:
#             magsuffix = " " + magsuffix

#     if appendBytes:
#         if appendBytes == "iec" and magsuffix:
#             bytesuffix = "iB" if base1024 else "B"
#         elif appendBytes == "short" and magsuffix:
#             bytesuffix = "B"
#         elif number == 1:
#             bytesuffix = " Byte"
#         else:
#             bytesuffix = " Bytes"

#     if thousandsSep and (number >= 1000 or magsuffix):
#         locale.setlocale(locale.LC_ALL, "")
#         # TODO: make precision configurable
#         if prec > 0:
#             fs = "%.{}f".format(prec)
#             snum = locale.format_string(fs, number, thousandsSep)
#         else:
#             snum = locale.format("%d", number, thousandsSep)
#         # Some countries like france use non-breaking-space (hex=a0) as group-
#         # seperator, that's not plain-ascii, so we have to replace the hex-byte
#         # "a0" with hex-byte "20" (space)
#         snum = hexlify(snum).replace("a0", "20").decode("hex")
#     else:
#         snum = str(number)

#     return "%s%s%s" % (snum, magsuffix, bytesuffix)


def lstrip_string(s, prefix, ignore_case=False):
    """Remove leading string from s.

    Note: This is different than `s.lstrip('bar')` which would remove
    all leading 'a', 'b', and 'r' chars.
    """
    if ignore_case:
        if s.lower().startswith(prefix.lower()):
            return s[len(prefix) :]
    else:
        if s.startswith(prefix):
            return s[len(prefix) :]
    return s


# def rstrip_string(s, suffix, ignore_case=False):
#     """Remove trailing string from s.

#     Note: This is different than `s.rstrip('bar')` which would remove
#     all trailing 'a', 'b', and 'r' chars.
#     """
#     if ignore_case:
#         if s.lower().endswith(suffix.lower()):
#             return s[:-len(suffix)]
#     else:
#         if s.endswith(suffix):
#             return s[:-len(suffix)]
#     return s



# -*- coding: utf-8 -*-
# (c) 2020-2023 Martin Wendt and contributors; see https://github.com/mar10/stressor
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
import re
import time
from copy import deepcopy

import requests
from snazzy import red, yellow

from stressor.config_manager import replace_var_macros
from stressor.context_stack import ContextStack
from stressor.plugins.base import ActivityAssertionError
from stressor.util import (
    NO_DEFAULT,
    StressorError,
    check_arg,
    get_dict_attr,
    logger,
    shorten_string,
)


class StoppedError(StressorError):
    """Raised when an activity stops due because the `stop_request` is set."""


class SkippedError(StressorError):
    """Raised when an activity is skipped due to `max_errors` or `max_time`."""


class User:
    def __init__(self, name, password, **kwargs):
        self.name = name
        self.password = password
        for arg_name, arg_val in kwargs.items():
            assert type(arg_name) in (int, float, str)
            setattr(self, arg_name, arg_val)
        return

    def __str__(self):
        return "User<{}>".format(self.name)

    # Provide nicer display for pprint(), etc.
    __repr__ = __str__

    @property
    def auth(self):
        """Return (name, password) tuple or None."""
        if self.password is None:
            return None
        return (self.name, self.password)


class SessionHelper:
    """Passed to script activities."""

    def __init__(self, session):
        self.__session = session

    @property
    def browser(self):
        return self.__session.browser_session

    def __repr__(self):
        return f"SessionHelper<{self.__session.session_id}>"


class SessionManager:
    """
    Run a scenario in a single session.
    """

    # #: (float)
    # DEFAULT_REQUEST_TIMEOUT = 10.0

    def __init__(self, run_manager, context, session_id, user):
        # check_arg(run_manager, RunManager)
        check_arg(context, dict)
        check_arg(session_id, str)
        check_arg(user, User, or_none=True)

        #: The :class:`RunManager` object that holds global settings and definitions
        self.run_manager = run_manager
        config = run_manager.config_manager.config

        # (dict) Global variables for this session. Initialized from the
        # run configuration, but not shared between sessions.
        # (`self.context` is accessible by the respective property below.)
        context = context.copy()
        #: (str) Unique ID string for this session
        self.session_id = session_id
        #: The :class:`User` object that is assigned to this session
        self.user = user or User("anonymous", "")
        #: (dict) Copy of `run_config.sessions` configuration
        self.sessions = run_manager.config_manager.sessions.copy()
        #: (dict) Activities can store per-session data here.
        #: Note that the activity objects are instintiated only once and shared
        #: by all sessions.
        self.data = {}
        #: (bool) True: only simulate activities
        self.dry_run = bool(context.get("dry_run"))
        #: (int) Verbosity 0..5
        self.verbose = context.get("verbose", 3)
        #: (:class:`threading.Event`)
        self.stop_request = run_manager.stop_request

        # Set some default entries in context dict
        # context.setdefault("timeout", self.DEFAULT_REQUEST_TIMEOUT)
        context.setdefault("session_id", self.session_id)
        context.setdefault("user", self.user)

        #: The :class:`~stressor.context_stack.ContextStack` object that reflects the current execution path
        self.context_stack = ContextStack(run_manager.host_id, context)
        self.context_stack.push(run_manager.process_id)
        self.context_stack.push(session_id)
        #: :class:`~stressor.statistic_manager.StatisticManager` object that containscurrent execution path
        self.stats = run_manager.stats
        # Lazy initialization using a property
        self._browser_session = None

        #: (int) Stop session if global error count > X
        #: Passing `--max-errors` will override this.
        self.max_errors = int(config.get("max_errors", 0))

        #: (float) Stop session if total run time  > X seconds.
        #: Passing `--max-time` will override this.
        self.max_time = float(config.get("max_time", 0.0))
        self._cancelled_seq = None

        # Used by StatisticsManager
        self.pending_sequence = None
        self.sequence_start = None
        self.pending_activity = None
        self.activity_start = None

        self.stats.register_session(self)

    def __str__(self):
        return "SessionManager<{}>".format(self.session_id)

    # Provide nicer display for pprint(), etc.
    __repr__ = __str__

    def publish(self, channel, *args, **kwargs):
        kwargs["session_id"] = self.session_id
        return self.run_manager.publish(channel, *args, **kwargs)

    def _evaluate_macros(self, kwargs, context):
        replace_var_macros(kwargs, context)
        return kwargs

    def make_session_helper(self):
        """Return a :class:`SessionHelper` instance for this session."""
        res = SessionHelper(self)
        return res

    @property
    def browser_session(self):
        """Return a ``requests.Session`` instance for this session."""
        if self._browser_session is None:
            self._browser_session = requests.Session()
        return self._browser_session

    @property
    def context(self):
        return self.context_stack.context

    @property
    def sess_stats(self):
        sess_stats = self.stats["sessions"][self.session_id]
        return sess_stats

    def get_context(self, dotted_key=None, default=NO_DEFAULT):
        res = self.context_stack.get_attr(dotted_key)
        return res

    def get_config(self, dotted_key=None, default=NO_DEFAULT):
        res = self.run_manager.config_manager.get(dotted_key, default)
        return res

    def log_info(self, *args):
        logger.info(self.session_id, *args)
        # self.publish("log", self.session_id, *args, level="info")

    def has_errors(self, or_warnings=False):
        return self.sess_stats["errors"] > 0

    def check_run_limits(self, seq_name):
        """Check if current time or number of errors exceeds the configured limits.

        Args:
            seq_name (str):
                current sequence name. Used to determine if a stopping session
                should still execute the 'end' sequence if the limit was reached
                during a main sequence.
        Returns:
            (bool) false if the current operation should be skipped.
        """
        cs = self._cancelled_seq
        err_limit_reached = (
            # Compare max_errors against global error count
            self.max_errors
            and self.stats.error_count() >= self.max_errors
        )

        run_time = self.run_manager.get_run_time()
        time_limit_reached = self.max_time and run_time > self.max_time

        if not cs and (err_limit_reached or time_limit_reached):
            # We just hit a limit: issue a warning and take a note of the sequence
            self._cancelled_seq = seq_name
            # self.stats.stats["run_limit_reached"] = True
            if err_limit_reached:
                msg = "Reached max. error limit of {}: stopping...".format(
                    self.max_errors
                )
            elif time_limit_reached:
                msg = "Reached max. run time limit of {}: stopping...".format(
                    self.max_time
                )
            self.stats.report_limit_violation(msg)
            logger.warning(yellow(msg))
            return False
        elif cs in ("init", "end"):
            # The limit was reached inside the init or end sequence: always skip
            return False
        elif cs:
            # The limit was reached inside a main sequence: allow to run the end
            # sequence but skip all others
            return seq_name == "end"
        return True

    def report_activity_start(self, sequence, activity):
        """Called by session runner before activities is executed."""
        path = self.context_stack.path()
        self.stats.report_start(self, sequence, activity, path=path)
        logger.info("{} {}".format("DRY-RUN" if self.dry_run else "Execute", path))

    def report_activity_error(self, sequence, activity, activity_args, exc):
        """Called session runner when activity `execute()` or assertions raise an error."""
        # self.stats.inc("errors")

        if isinstance(exc, SkippedError):
            logger.warning(yellow("Skipped {}".format(activity)))
            self.pending_activity = None
            return

        # Create a copy of the current context, so we can shorten values
        context = self.context_stack.context.copy()
        context["last_result"] = shorten_string(context.get("last_result"), 500, 100)

        msg = []
        # msg.append("{} {}: {!r}:".format(self.context_stack, activity, exc))
        # msg.append("{!r}:".format(exc))
        msg.append("{!r}:".format(exc))
        if isinstance(exc, ActivityAssertionError):
            msg.append("Failed assertions:")
            for err in exc.assertion_list:
                msg.append("  - {}".format(err))
        msg.append("Execution path: {}".format(self.context_stack))
        msg.append("Activity: {}".format(activity))
        msg.append("Activity args: {}".format(activity_args))
        msg.append("Context: {}".format(context))

        msg = "\n    ".join(msg)
        logger.error(red(msg))
        self.stats.report_error(self, sequence, activity, error=msg)
        return

    def report_activity_result(self, sequence, activity, activity_args, result, elap):
        """Called session runner when activity `execute()` completes."""
        self.stats.report_end(self, sequence, activity)

    def _process_activity_result(self, activity, activity_args, result, elap):
        """Perform common checks.

        Raises:
            ActivityAssertionError
        """
        context = self.context_stack.context
        errors = []
        # warnings = []

        arg = float(activity_args.get("assert_max_time", 0))
        if arg and elap > arg:
            errors.append(
                "Execution time limit of {} seconds exceeded: {:.3} sec.".format(
                    arg, elap
                )
            )

        arg = activity_args.get("assert_match")
        if arg:
            text = str(result)
            # Note: use re.search (not .match)!
            if not re.search(arg, text, re.MULTILINE):
                errors.append(
                    "Result does not match `{}`: {!r}".format(
                        arg, shorten_string(text, 500, 100)
                    )
                )

        arg = activity_args.get("store_json")
        if arg:
            for var_name, key_path in arg.items():
                try:
                    val = get_dict_attr(result, key_path)
                    context[var_name] = val
                except Exception:
                    errors.append(
                        "store_json could not find `{}` in activity result {!r}".format(
                            key_path, result
                        )
                    )

        if errors:
            raise ActivityAssertionError(errors)
        return

    def run_sequence(self, seq_name, sequence):
        stack = self.context_stack
        context = stack.context

        self.publish(
            "start_sequence",
            session=self,
            sequence=sequence,
            path=stack,
        )
        self.stats.report_start(self, seq_name, None)
        start_sequence = time.monotonic()
        for act_idx, activity_args in enumerate(sequence, 1):
            # activity_args["activity"] is an instance of ActivityBase that
            # we want to re-use it for every session.
            # The rest of activity_args is copied, so session data is separated
            activity = activity_args["activity"]
            activity_args = deepcopy(activity_args)
            activity_args.pop("activity")

            # Add activity info to path
            # Note: `get_info()` is not as detailed as it could, since we don't
            # pass the expanded args here. We set it anyway, so we have a valid
            # stack in case `_evaluate_macros()` blows.
            with stack.enter(
                "#{:02}-{}".format(act_idx, activity.get_info(session=self))
            ):
                expanded_args = self._evaluate_macros(activity_args, context)

                # Let activity do internal calculations, that might be used by
                # the follwing call to `get_info()`
                activity.prepare_execute(self, expanded_args)

                # Enhance the path info with expanded args
                stack.set_last_part(
                    activity.get_info(expanded_args=expanded_args, session=self)
                )

                error = None
                result = None
                self.publish(
                    "start_activity",
                    session=self,
                    sequence=sequence,
                    activity=activity,
                    expanded_args=expanded_args,
                    context=context,
                    path=stack,
                )
                start_activity = time.monotonic()

                self.report_activity_start(seq_name, activity)

                try:
                    if self.stop_request.is_set():
                        raise StoppedError
                    if not self.check_run_limits(seq_name):
                        raise SkippedError

                    result = activity.execute(self, **expanded_args)
                    context["last_result"] = result
                    # Evaluate standard `assert_...` and `store_...` clauses:
                    elap = time.monotonic() - start_activity
                    self._process_activity_result(
                        activity,
                        activity_args,
                        result,
                        elap,
                    )
                    self.report_activity_result(
                        seq_name,
                        activity,
                        activity_args,
                        result,
                        elap,
                    )
                except (Exception, KeyboardInterrupt) as e:
                    if isinstance(e, KeyboardInterrupt):
                        self.stop_request.set()
                    error = e
                    self.report_activity_error(seq_name, activity, activity_args, e)
                    if not isinstance(e, (KeyboardInterrupt, StressorError)):
                        logger.exception("")
                    # return False

                finally:
                    elap = time.monotonic() - start_activity
                    self.publish(
                        "end_activity",
                        session=self,
                        sequence=sequence,
                        path=stack,
                        activity=activity,
                        result=result,
                        error=error,
                        elap=elap,
                        context=context,
                    )

        elap = time.monotonic() - start_sequence
        self.stats.report_end(self, seq_name, None)
        self.publish(
            "end_sequence",
            session=self,
            sequence=sequence,
            path=stack,
            elap=elap,
        )
        context["last_result"] = None
        return not self.has_errors()

    def run(self):
        stack = self.context_stack
        rm = self.run_manager
        config_manager = rm.config_manager
        config = config_manager.config
        sequences = config_manager.sequences
        scenario = config_manager.scenario
        sessions = config_manager.sessions
        session_duration = float(sessions.get("duration", 0.0))

        self.publish("start_session", session=self)
        self.stats.report_start(self, None, None)

        start_session = time.monotonic()
        skip_all = False
        skip_all_but_end = False

        for seq_idx, seq_def in enumerate(scenario, 1):
            seq_name = seq_def["sequence"]
            if skip_all or (skip_all_but_end and seq_name != "end"):
                logger.warning("Skipping sequence '{}'.".format(seq_name))
                continue

            sequence = sequences.get(seq_name)
            loop_repeat = int(seq_def.get("repeat", 0))
            loop_duration = float(seq_def.get("duration", 0.0))
            start_seq_loop = time.monotonic()
            loop_idx = 0
            while True:
                loop_idx += 1
                if not self.check_run_limits(seq_name=seq_name):
                    skip_all_but_end = True
                    break
                # One single pass by default
                if not loop_repeat and not loop_duration and loop_idx > 1:
                    break
                # `Sequence repeat: COUNT`:
                if loop_repeat and loop_idx > loop_repeat:
                    break
                # `--single`:
                if loop_idx > 1 and config.get("force_single"):
                    logger.warning(
                        "force_single: sequence '{}' skipping remaining {} loops.".format(
                            seq_name, loop_repeat - 1 if loop_repeat else ""
                        )
                    )
                    break

                now = time.monotonic()
                # `Sequence duration: SECS`:
                if loop_duration > 0 and now > (start_seq_loop + loop_duration):
                    logger.info(
                        "Stopping sequence '{}' loop after {} sec.".format(
                            seq_name, loop_duration
                        )
                    )
                    break
                # `Session duration: SECS` (but run 'end' sequence):
                elif (
                    seq_name != "end"
                    and session_duration > 0
                    and now > (start_session + session_duration)
                ):
                    logger.info(
                        "Stopping scenario '{}' loop after {} sec.".format(
                            seq_name, session_duration
                        )
                    )
                    skip_all_but_end = True
                    break

                with stack.enter("#{:02}-{}@{}".format(seq_idx, seq_name, loop_idx)):
                    is_ok = self.run_sequence(seq_name, sequence)
                    if seq_name == "init" and not is_ok:
                        logger.error(
                            "Stopping scenario due to an error in the 'init' sequence."
                        )
                        skip_all = True
                        break
                    elif self.stop_request.is_set():
                        logger.error("Stopping scenario due to a stop request.")
                        # TODO: a second 'ctrl-c' should not be so graceful
                        skip_all_but_end = True
                        break
            # self.stats.report_end(self, seq_name, None)

        elap = time.monotonic() - start_session
        self.stats.report_end(self, None, None)

        self.publish("end_session", session=self, elap=elap)

        # if self._cancelled_seq:
        #     return False
        return not self.has_errors()



# -*- coding: utf-8 -*-
# (c) 2020-2023 Martin Wendt and contributors; see https://github.com/mar10/stressor
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""

import os
import tempfile

import pytest

from stressor.convert.har_converter import HarConverter


class TestConvert:
    def setup_method(self):
        self.fixtures_path = os.path.join(os.path.dirname(__file__), "fixtures")

    def teardown_method(self):
        pass

    def test_1(self):
        with tempfile.TemporaryDirectory() as target_folder:
            opts = {
                "fspec": os.path.join(self.fixtures_path, "har_1.har"),
                "target_folder": target_folder,
                "collate_max_len": 0,
            }
            conv = HarConverter(opts)
            conv.run()
            act_yaml_path = os.path.join(target_folder, "main_sequence.yaml")
            yaml = open(act_yaml_path, "rt").read()
            # print(yaml)
            assert len(conv.entries) == 5
            assert conv.har_version == "1.2"
        assert "activity: GetRequest" in yaml
        assert "url: '/test1.json'" in yaml
        # assert 'url: "http://127.0.0.1:8082/test1.json"' in yaml
        # assert 0

    def test_2(self):
        # This test is only useful to inspect the folder while developing
        # TODO: remove later
        target_folder = "/Users/martin/temp"
        if not os.path.isdir(target_folder):
            pytest.skip("Local test folder not found")
        opts = {
            "fspec": os.path.join(self.fixtures_path, "har_2.har"),
            "target_folder": target_folder,
            "force": True,
        }
        conv = HarConverter(opts)
        conv.run()
        # yaml = open(opts["fspec"], "rt").read()
        # print(yaml)
        assert len(conv.entries) == 46
        assert conv.stats["entries"] == 46
        assert conv.stats["entries_total"] == 81
        assert conv.stats["external_urls"] == 6
        assert conv.stats["skipped"] == 35
        assert conv.stats["collated_activities"] == 9
        assert conv.stats["collated_urls"] == 41

        assert conv.har_version == "1.2"
        # assert 0



# -*- coding: utf-8 -*-
# (c) 2020-2023 Martin Wendt and contributors; see https://github.com/mar10/stressor
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
import os

import pytest

from stressor.config_manager import (
    ConfigManager,
    ConfigurationError,
    replace_var_macros,
)
from stressor.plugin_manager import PluginManager
from stressor.statistic_manager import StatisticManager


class TestConfigReader:
    def setup_method(self):
        self.fixtures_path = os.path.join(os.path.dirname(__file__), "fixtures")
        PluginManager.register_plugins(arg_parser=None)

    def test_path_check(self):
        path = os.path.join(self.fixtures_path, "INVALID")
        stats = StatisticManager()
        cr = ConfigManager(stats)
        with pytest.raises(ConfigurationError, match=r".*not found.*"):
            cr.read(path)

    def test_replace_var_macros(self):
        context = {
            "root_url": "http://example.com/$(mountpoint)",
            "mountpoint": "s1",
            "guid": "42",
            "user": {"name": "joe", "password": "secret"},
        }
        value = {
            "a": "$(root_url)?guid=$(guid)",
            "b": "$(root_url)?user=$(user.name)",
        }
        replace_var_macros(value, context)
        assert value["a"] == "http://example.com/s1?guid=42"
        assert value["b"] == "http://example.com/s1?user=joe"

    def test_read_scenario(self):
        path = os.path.join(self.fixtures_path, "test_dry_run")
        stats = StatisticManager()
        cm = ConfigManager(stats)
        res = cm.read(path)
        assert res is cm.config_all
        assert cm.root_folder == self.fixtures_path

        # Check load time macros
        assert cm.sessions["users"][0] == {
            "name": "User_1",
            "password": "secret",
            "custom": "foo",
        }, "$load() macro"

        # SleepActivity
        activity_dict = cm.sequences["init"][1]
        assert activity_dict["activity"].__class__.__name__ == "SleepActivity"
        assert activity_dict["duration"] == 0.01

        # $sleep(.01):
        activity_dict = cm.sequences["init"][2]
        assert activity_dict["activity"].__class__.__name__ == "SleepActivity"
        assert activity_dict["duration"] == 0.02

        # $sleep($(default_sleep))
        activity_dict = cm.sequences["init"][3]
        assert activity_dict["activity"].__class__.__name__ == "SleepActivity"
        assert activity_dict["duration"] == "$(default_sleep)"

        # $sleep(0.01, .02):
        activity_dict = cm.sequences["init"][4]
        assert activity_dict["activity"].__class__.__name__ == "SleepActivity"
        assert activity_dict["duration"] == 0.01
        assert activity_dict["duration_2"] == 0.02

        return

    # def test_read_scenario_2(self):
    #     path = os.path.join(self.fixtures_path, "test_mock_server")
    #     cr = ConfigManager(None)
    #     res = cr.read(path)
    #     assert res is cr.config_all
    #     assert cr.root_folder == self.fixtures_path
    #     # Check load time macros
    #     assert cr.run_config["sessions"]["users"][0] == {
    #         "name": "User_1",
    #         "password": "secret",
    #     }, "$load() macro"

    #     activity_dict = cr.sequences["init"][2]
    #     assert activity_dict["activity"].__class__.__name__ == "SleepActivity"
    #     assert activity_dict["duration"] == ".1"

    #     return



# -*- coding: utf-8 -*-
# (c) 2020-2023 Martin Wendt and contributors; see https://github.com/mar10/stressor
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
from stressor.plugin_manager import PluginManager


class TestPluginManager:
    # def setup_method(self):
    #     self.fixtures_path = os.path.join(os.path.dirname(__file__), "fixtures")

    def test_plugins(self):
        pm = PluginManager
        pm.register_plugins(arg_parser=None)
        assert "PsAlloc" in pm.activity_plugin_map
